"""Functions for multi-dimensional images."""
import numpy as np
from tqdm import tqdm
from scipy import ndimage

from psdist.utils import array_like
from psdist.utils import cov2corr
from psdist.utils import edges_from_centers
from psdist.utils import centers_from_edges


# Analysis
# ------------------------------------------------------------------------------


def get_grid_coords(*coords):
    """Return list of grid coordinates from coordinate arrays along each axis.

    Parameters
    ----------
    coords : list[ndarray]
        Coordinates along each axis of regular grid. Example: [[1, 2, 3], [0, 1, 2]].

    Returns
    -------
    ndarray, shape (K, len(coords))
        Coordinate array for all points in the grid. The total number of grid
        points is `K = np.prod([len(c) for c in coords])`.
    """
    return np.vstack([C.ravel() for C in np.meshgrid(*coords, indexing="ij")]).T


def max_indices(f):
    """Return the indices of the maximum element of `f`."""
    return np.unravel_index(np.argmax(f), f.shape)


def get_radii(coords, Sigma):
    """Return "radii" (x^T Sigma^-1^T x) from grid coordinates and covariance matrix.

    This is quite slow when d > 4 due to creating a mesh grid.

    Parameters
    ----------
    coords : list[ndarray], length d
        Coordinate array for each dimension of the regular grid.
    Sigma : ndarray, shape (d, d)
        Covariance matrix of some distribution on the grid.

    Returns
    -------
    R : ndarray
        "Radius" x^T Sigma^-1^T x at each point in grid.
    """
    COORDS = np.meshgrid(*coords, indexing="ij")
    shape = tuple([len(c) for c in coords])
    R = np.zeros(shape)
    Sigma_inv = np.linalg.inv(Sigma)
    for ii in tqdm(np.ndindex(shape)):
        vec = np.array([C[ii] for C in COORDS])
        R[ii] = np.sqrt(np.linalg.multi_dot([vec.T, Sigma_inv, vec]))
    return R


def radial_density(f, R, radii, dr=None):
    """Return average density within ellipsoidal shells.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    R : ndarray, same shape as `f`.
        Gives the "radius" at each pixel in f.
    radii : ndarray, shape (k,)
        Radii at which to evaluate the density.
    dr : float
        The shell width.

    Returns
    -------
    fr : ndarray, shape (k,)
        The average density within each ellipsoidal shell.
    """
    if dr is None:
        dr = 0.5 * np.max(R) / (len(R) - 1)
    fr = []
    for r in tqdm(radii):
        f_masked = np.ma.masked_where(np.logical_or(R < r, R > r + dr), f)
        # mean density within this shell...
        fr.append(np.mean(f_masked))
    return np.array(fr)


def mean(f, coords=None):
    """Compute the d-dimensional mean.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinates along each axis of the image.

    Returns
    -------
    ndarray, shape (d,)
        The d-dimensional mean.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]
    return np.array(
        [np.average(C, weights=f) for C in np.meshgrid(*coords, indexing="ij")]
    )


def cov(f, coords=None):
    """Compute the d x d covariance matrix.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinates along each axis of the image.

    Returns
    -------
    ndarray, shape (d, d)
        The covariance matrix.
    """

    def cov_2x2(_f, _coords):
        COORDS = np.meshgrid(*_coords, indexing="ij")
        Sigma = np.zeros((_f.ndim, _f.ndim))
        _f_sum = np.sum(_f)
        if _f_sum > 0:
            mean = np.array([np.average(C, weights=_f) for C in COORDS])
            for i in range(_f.ndim):
                for j in range(i + 1):
                    X = COORDS[i] - mean[i]
                    Y = COORDS[j] - mean[j]
                    EX = np.sum(_f * X) / _f_sum
                    EY = np.sum(_f * Y) / _f_sum
                    EXY = np.sum(_f * X * Y) / _f_sum
                    Sigma[i, j] = Sigma[j, i] = EXY - EX * EY
        return Sigma

    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]

    if f.ndim < 3:
        return cov_2x2(f, coords)

    Sigma = np.zeros((f.ndim, f.ndim))
    for i in range(f.ndim):
        for j in range(i):
            axis = (i, j)
            _image = project(f, axis=axis)
            _coords = [coords[k] for k in axis]
            # Compute 2 x 2 covariance matrix from this projection.
            _sigma = cov_2x2(_image, _coords)
            # Update elements of n x n covariance matrix. This will update
            # some elements multiple times, but it should not matter.
            Sigma[i, i] = _sigma[0, 0]
            Sigma[j, j] = _sigma[1, 1]
            Sigma[i, j] = Sigma[j, i] = _sigma[0, 1]
    return Sigma


def corr(f, coords=None):
    """Compute the d x d correlation matrix.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinates along each axis of the image.

    Returns
    -------
    ndarray, shape (d, d)
        The correlation matrix.
    """
    return cov2corr(cov(f, coords))


# Transformation
# ------------------------------------------------------------------------------


def slice_idx(d=1, axis=0, ind=0):
    """Return planar slice index array.

    Parameters
    ----------
    d : int
        The number of elements in the slice index array. (The number of dimensions
        in the array to be sliced.)
    axis : list[int]
        The sliced axes.
    ind : list[int] or list[tuple]
        The indices along the sliced axes. If a tuple is provided, this
        defines the (min, max) index.

    Returns
    -------
    idx : n-tuple
        The slice index array. A slice of the array `f` may then be accessed as
        `f[idx]`.
    """
    # Make list if only one axis provided.
    if type(axis) is int:
        axis = [axis]
        # Can also provide only one axis but provide a tuple for ind, which
        # selects a range along that axis.
        if type(ind) is tuple:
            ind = [ind]
    # Make list if only one ind provided.
    if type(ind) is int:
        ind = [ind]
    # Initialize the slice index to select all elements.
    idx = d * [slice(None)]
    # If any indices were provided, add them to `idx`.
    for k, item in zip(axis, ind):
        if item is None:
            continue
        elif type(item) is tuple and len(item) == 2:
            idx[k] = slice(item[0], item[1])
        else:
            # Could be int or list of ints
            idx[k] = item
    return tuple(idx)


def slice_idx_ellipsoid(f, axis=None, rmin=0.0, rmax=1.0):
    """Compute an ellipsoid slice.

    Ellipsoid is computed from the covariance matrix of `f`.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    axis : list[int]
        Specificies the subspace in which the ellipsoid slices are computed.
        Example: in x-y-z space, we may define a circle in x-y. This could
        select points within a cylinder in x-y-z.
    rmin, rmax : float
        We select the region between two nested ellipsoids with "radius"
        rmin and rmax. The radius is r = x^T Sigma^-1 x, where Sigma is
        the covariance matrix and x is the coordinate vector. r = 1 is
        the covariance ellipsoid.

    Returns
    -------
    np.ma.masked_array
        A version of `f` in which elements outside the slice are masked.
    """
    # Will need to compute an (n-m)-dimensional mask (m = len(axis)), then
    # copy the mask into the remaining dimensions with `copy_into_new_dim`.
    raise NotImplementedError


def slice_idx_contour(f, axis=None, lmin=0.0, lmax=1.0):
    """Compute a contour slice.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    axis : list[int]
        Specificies the subspace in which the contours are computed. (See
        `slice_idx_ellipsoid`.)
    lmin, lmax : float
        `f`is projected onto `axis` and the projection `fpr` is normalized to
        the range [0, 1]. Then, we find the points in this subspace such that
        `fpr` is within the range [lmin, lmax].

    Returns
    -------
    np.ma.masked_array
        A version of `f` in which elements outside the slice are masked.
    """
    # Will need to compute an (n-m)-dimensional mask (m = len(axis)), then
    # copy the mask into the remaining dimensions with `copy_into_new_dim`.
    raise NotImplementedError


def _slice(f, axis=0, ind=0):
    idx = slice_idx(f.ndim, axis=axis, ind=ind)
    return f[idx]


def _slice_ellipsoid(f, axis=None, rmin=0.0, rmax=1.0):
    idx = slice_idx_ellipsoid(f, axis=axis, rmin=rmin, rmax=rmax)


def _slice_contour(f, axis=None, lmin=0.0, lmax=1.0):
    idx = slice_idx_contour(f, axis=axis, lmin=lmin, lmax=lmax)
    return f[idx]


def project(f, axis=0):
    """Project along one or more axes.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    axis : list[int]
        The axes onto which the image is projected, i.e., the
        axes which are not summed over. Can be an int or list
        of ints. Array axes are swapped as required.

    Returns
    -------
    proj : ndarray
        The projection of `image` onto the specified axis.
    """
    # Sum over specified axes.
    if type(axis) is int:
        axis = [axis]
    axis = tuple(axis)
    axis_sum = tuple([i for i in range(f.ndim) if i not in axis])
    proj = np.sum(f, axis_sum)

    # Order the remaining axes.
    loc = list(range(proj.ndim))
    destination = np.zeros(proj.ndim, dtype=int)
    for i, index in enumerate(np.argsort(axis)):
        destination[index] = i
    for i in range(proj.ndim):
        if loc[i] != destination[i]:
            j = loc.index(destination[i])
            proj = np.swapaxes(proj, i, j)
            loc[i], loc[j] = loc[j], loc[i]
    return proj


def project1d_contour(f, axis=0, lmin=0.0, lmax=1.0, fpr=None):
    """Apply contour slice in d - 1 dimensions, then project onto the remaining dimension.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    axis : int
        The projection axis.
    lmin, lmax : float
        Min and max contour levels of the (d-1)-dimensional projection of `f`,
        normalized the the range [0, 1].
    fpr : ndarray, shape [f.shape[i] for i in range(f.ndim) if i != axis]
        The (d-1)-dimensional projection of `f` onto all dimensions other than `axis`.
        (If not provided, it will be computed within the function.)

    Returns
    -------
    ndarray, shape (f.shape[axis],)
        The projection of the slice.
    """
    axis_proj = [i for i in range(f.ndim) if i != axis]
    if fpr is None:
        fpr = project(f, axis=axis_proj)
    fpr = fpr / np.max(fpr)
    idx = slice_idx(
        d=f.ndim,
        axis=axis_proj,
        ind=np.where(np.logical_and(fpr >= lmin, fpr <= lmax)),
    )
    # `f[idx]` will give a two-dimensional array. Normally we need to sum over
    # the first axis. If `axis == 0`, we need to sum over the second axis.
    return np.sum(f[idx], axis=int(axis == 0))


def project2d_contour(f, axis=(0, 1), lmin=0.0, lmax=1.0, fpr=None):
    """Apply contour slice in d - 2 dimensions, then project onto the remaining two dimensions.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    axis : tuple
        The 2D projection axis.
    lmin, lmax : float
        Min and max contour levels of the (n-2)-dimensional projection of `f`,
        normalized the the range [0, 1].
    fpr : ndarray, shape [f.shape[i] for i in range(f.ndim) if i != axis]
        The (n-1)-dimensional projection of `f` onto all dimensions other than `axis`.
        (If not provided, it will be computed within the function.)

    Returns
    -------
    ndarray, shape (f.shape[axis[0]], f.shape[axis[1]])
        The 2D projection of the slice.
    """
    axis_proj = [k for k in range(f.ndim) if k not in axis]
    if fpr is None:
        fpr = project(f, axis=axis_proj)
    fpr = fpr / np.max(fpr)
    idx = slice_idx(
        f.ndim, axis_proj, np.where(np.logical_and(fpr >= lmin, fpr <= lmax))
    )
    # `f[idx]` will give a three-dimensional array. Normally we need to sum over
    # the first axis. If `axis == (0, 1)`, we need to sum over the third axis.
    # If `axis == (0, n - 1), we need to sum over the second axis.
    _axis_proj = (1, 2)
    if axis == (0, 1):
        _axis_proj = (0, 1)
    elif axis == (0, f.ndim - 1):
        _axis_proj = (0, 2)
    # Two elements of `idx` will be `slice(None)`; these are the elements in `axis`.
    # These will always be in order. So, if `axis[0] > axis[1]`, we need to flip
    # `axis_proj`. Need a way to handle this automatically.
    if axis[0] > axis[1]:
        _axis_proj = tuple(reversed(_axis_proj))
    return project(f[idx], axis=_axis_proj)


def copy_into_new_dim(f, shape=None, axis=-1, method="broadcast", copy=False):
    """Copy image into one or more new dimensions.

    See 'https://stackoverflow.com/questions/32171917/how-to-copy-a-2d-array-into-a-3rd-dimension-n-times'

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    shape : d-tuple of ints
        The shape of the new dimensions.
    axis : int (0 or -1)
        If 0, the new dimensions will be inserted before the first axis. If -1,
        the new dimensions will be inserted after the last axis. I think
        values other than 0 or -1 should work; this does not currently
        work, at least for `method='broadcast'`, last I checked.
    method : {'repeat', 'broadcast'}
        Whether to use `np.repeat` or `np.expand_dims` and `np.broadcast_to`. The
        'broadcast' method is faster.
    """
    if not array_like(shape):
        shape = (shape,)
    if method == "repeat":
        for i in range(len(shape)):
            f = np.repeat(np.expand_dims(f, axis), shape[i], axis=axis)
        return f
    elif method == "broadcast":
        if axis == 0:
            new_shape = shape + f.shape
        elif axis == -1:
            new_shape = f.shape + shape
        else:
            raise ValueError("Cannot yet handle axis != 0, -1.")
        for _ in range(len(shape)):
            f = np.expand_dims(f, axis)
        if copy:
            return np.broadcast_to(f, new_shape).copy()
        else:
            return np.broadcast_to(f, new_shape)
    return None


# Processing

def blur(f, sigma):
    """Simple Gaussian blur."""
    return ndimage.gaussian_filter(f, sigma)


def clip(f, lmin=None, lmax=None, frac=False):
    """Clip between lmin and lmax, can be fractions or absolute values."""
    if not (lmin or lmax):
        return f
    if frac:
        f_max = np.max(f)
        if lmin:
            lmin = f_max * lmin
        if lmax:
            lmax = f_max * lmax
    return np.clip(f, lmin, lmax)


def fill(f, fill_value=None):
    """Fill masked values."""
    return np.ma.filled(f, fill_value=fill_value)


def normalize(f, norm="volume", pixel_volume=1.0):
    """Scale to unit volume or unit maximum."""
    factor = 1.0
    if norm == "volume":
        factor = np.sum(f) * pixel_volume
    elif norm == "max":
        factor = np.max(f)
    if factor == 0.0:
        return f
    return f / factor


def threshold(f, lmin=None, frac=False):
    """Set pixels less than lmin to zero."""
    if lmin:
        if frac:
            f_max = np.max(f)
            lmin = lmin * f_max
        f[f < lmin] = 0.0
    return f


def process(
    f,
    fill_value=None,
    thresh=None,
    thresh_type="abs",
    clip=None,
    clip_type="abs",
    norm=None,
    pixel_volume=1.0,
    blur_sigma=None,
):
    """Return processed image.

    This is a convenience function mostly for plotting routines.

    Parameters
    ----------
    f : ndarray
        A two-dimensional image.
    fill_value : float
        Fill masked elements of `f` with this value.
    mask_nonpositive : bool
        Masks mask non-positive values of `f`.
    thresh : float
        Set elements below this value to zero.
    clip: (lmin, lmax)
        Clip (limit) elements to within the range [lmin, lmax].
    thresh_type, clip_type : {'abs', 'frac'}
        Whether `thresh` and `clip` refer to absolute values or fractions
        of the maximum element of `f`.
    norm : {None, 'max', 'volume'}
        Whether to normalize the image by its volume or maximum element.
    pixel_volume : float
        Needed if normalizing by volume.
    blur_sigma : float
        Sigma for Gaussian filter.
    """
    if fill_value is not None:
        f = fill(f, fill_value=None)
    if thresh is not None:
        f = threshold(f, thresh, frac=(thresh_type == "frac"))
    if clip is not None:
        f = clip(f, clip[0], clip[1], frac=(clip_type == "frac"))
    if blur_sigma is not None:
        f = blur(f, blur_sigma)
    if norm:
        f = normalize(f, norm=norm, pixel_volume=pixel_volume)
    return f


# Sampling
# ------------------------------------------------------------------------------


def sample_hist(f, coords=None, samples=1):
    """Sample from histogram.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image/histogram.
    coords : list[ndarray]
        Coordinates (bin centers) along each axis.
    samples : int
        The number of samples to draw.

    Returns
    -------
    ndarray, shape (samples, d)
        Samples drawn from the distribution.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]
    elif f.ndim == 1:
        coords = [coords]
    edges = [edges_from_centers(c) for c in coords]

    idx = np.flatnonzero(f)
    pdf = f.ravel()[idx]
    pdf = pdf / np.sum(pdf)
    idx = np.random.choice(idx, samples, replace=True, p=pdf)
    idx = np.unravel_index(idx, shape=f.shape)
    lb = [edges[axis][idx[axis]] for axis in range(f.ndim)]
    ub = [edges[axis][idx[axis] + 1] for axis in range(f.ndim)]
    return np.squeeze(np.random.uniform(lb, ub).T)


def sample_sparse_hist(indices=None, counts=None, coords=None, samples=1):
    """Sample from sparse histogram.

    Parameters
    ----------
    indices : ndarray, shape (k, d)
        d-dimensional histogram bin indices.
    counts : ndarray, shape (k,)
        Counts in each bin. Does not need to be normalized.
    coords : list[ndarray]
        Coordinates (bin centers) along each axis.        
    samples : int
        The number of samples to draw.

    Returns
    -------
    ndarray, shape (samples, d)
        Samples drawn from the distribution.
    """
    shape = [len(c) for c in coords]
    edges = [edges_from_centers(c) for c in coords]
    indices_flat = np.ravel_multi_index(indices.T, shape)
    idx = np.random.choice(indices_flat, size=samples, replace=True, p=(counts / np.sum(counts)))
    idx = np.unravel_index(idx, shape=shape)
    lb = [edges[axis][idx[axis]] for axis in range(len(shape))]
    ub = [edges[axis][idx[axis] + 1] for axis in range(len(shape))]
    return np.squeeze(np.random.uniform(lb, ub).T)


import numpy as np
import proplot as pplt

import psdist.image
import psdist.cloud
import psdist.utils
import psdist.visualization.cloud as vis_cloud
import psdist.visualization.image as vis_image
from psdist.visualization.visualization import plot1d


class JointGrid:
    """Grid for joint plots.

    https://seaborn.pydata.org/generated/seaborn.JointGrid.html

    Attributes
    -----------
    fig : proplot.figure.Figure
        The main figure.
    ax : proplot.gridspec.SubplotGrid
        The main axis.
    ax_panel_x, ax_panel_y : proplot.gridspec.SubplotGrid
        The panel (marginal) axes on the top and right.
    """

    def __init__(
        self, panel_kws=None, panel_fmt_kws_x=None, panel_fmt_kws_y=None, **fig_kws
    ):
        """Constructor.

        marg_kws : dict
            Key word arguments for `ax.panel`.
        marg_fmt_kws_x, marg_fmt_kws_y : dict
            Key word arguments for `ax.format` for each of the panel axs.
        **fig_kws
            Key word arguments passed to `proplot.subplots`.
        """
        self.fig, self.ax = pplt.subplots(**fig_kws)

        if panel_kws is None:
            panel_kws = dict()
        panel_kws.setdefault("space", 0.0)
        self.ax_panel_x = self.ax.panel("t", **panel_kws)
        self.ax_panel_y = self.ax.panel("r", **panel_kws)
        self.panel_axs = [self.ax_panel_x, self.ax_panel_y]
        if panel_fmt_kws_x is None:
            panel_fmt_kws_x = dict()
        if panel_fmt_kws_y is None:
            panel_fmt_kws_y = dict()
        panel_fmt_kws = [panel_fmt_kws_x, panel_fmt_kws_y]
        for i, key in enumerate(["ylim", "xlim"]):
            panel_fmt_kws[i].setdefault(key, (0.0, 1.1))
        for ax, kws in zip(self.panel_axs, panel_fmt_kws):
            kws.setdefault("xspineloc", "neither")
            kws.setdefault("yspineloc", "neither")
            ax.format(**kws)

    def plot_cloud(self, X, marg_hist_kws=None, marg_kws=None, **kws):
        """Plot a 2D point cloud.

        Parameters
        ----------
        X : ndarray, shape (k, 2)
            Coordinates of k points in 2-dimensional space.
        marg_hist_kws : dict
            Key word arguments passed to `np.histogram` for 1D histograms.
        marg_kws : dict
            Key word arguments passed to `visualization.plot1d`.
        **kws
            Key word arguments passed to `visualization.image.plot2d.`
        """
        if marg_kws is None:
            marg_kws = dict()
        marg_kws.setdefault("kind", "step")
        marg_kws.setdefault("lw", 1.0)
        if marg_hist_kws is None:
            marg_hist_kws = dict()
        marg_hist_kws.setdefault("bins", "auto")
        kws.setdefault("kind", "hist")
        if kws["kind"] == "hist":
            kws.setdefault("mask", True)
            kws.setdefault("bins", marg_hist_kws["bins"])
        if kws["kind"] != "scatter":
            kws.setdefault("colorbar_kw", dict())
            kws["colorbar_kw"].setdefault("pad", 2.0)
        for axis in range(2):
            profile, edges = np.histogram(X[:, axis], **marg_hist_kws)
            profile = profile / np.max(profile)
            centers = psdist.utils.centers_from_edges(edges)
            plot1d(
                centers, profile, ax=self.panel_axs[axis], flipxy=bool(axis), **marg_kws
            )
        vis_cloud.plot2d(X, ax=self.ax, **kws)

    def plot_image(self, f, coords=None, marg_kws=None, **kws):
        """Plot a 2D image.

        Parameters
        ----------
        f : ndarray
            A d-dimensional image.
        coords : list[ndarray]
            Coordinates along each dimension of `f`.
        marg_kws : dict
            Key word arguments passed to `visualization.plot1d`.
        **kws
            Key word arguments passed to `visualization.image.plot2d.`
        """
        kws.setdefault("colorbar_kw", dict())
        kws["colorbar_kw"].setdefault("pad", 2.0)
        if marg_kws is None:
            marg_kws = dict()
        marg_kws.setdefault("color", "black")
        marg_kws.setdefault("kind", "step")
        marg_kws.setdefault("lw", 1.0)
        if coords is None:
            coords = [np.arange(f.shape[axis]) for axis in range(f.ndim)]
        _out = vis_image.plot2d(f, coords=coords, ax=self.ax, **kws)
        fx = psdist.image.project(f, axis=0)
        fy = psdist.image.project(f, axis=0)
        fx = fx / np.max(fx)
        fy = fy / np.max(fy)
        for axis in range(2):
            profile = psdist.image.project(f, axis)
            profile = profile / np.max(profile)
            plot1d(
                coords[axis],
                profile,
                ax=self.panel_axs[axis],
                flipxy=bool(axis),
                **marg_kws,
            )
        return _out

    def colorbar(self, mappable, **kws):
        """Add a colorbar."""
        kws.setdefault("loc", "r")
        kws.setdefault("pad", 2.0)
        self.fig.colorbar(mappable, **kws)


class CornerGrid:
    """Grid for corner plots.

    * https://seaborn.pydata.org/generated/seaborn.PairGrid.html
    * https://corner.readthedocs.io/en/latest/
    * https://pandas.pydata.org/docs/reference/api/pandas.plotting.scatter_matrix.html

    Attributes
    ----------
    fig : proplot.figure.Figure
        The main figure.
    axs : proplot.gridspec.SubplotGrid
        The subplot axes.
    diag_axs : list[proplot.gridspec.SubplotGrid]
        The axes for diagonal (univariate) subplots. Can be empty.
    offdiag_axs : list[proplot.gridspec.SubplotGrid]
        The axes for off-diagonal (bivariate) subplots.
    diag_indices : list[int]
        The index of the dimension plotted on each diagonal subplot.
    offdiag_indices : list[2-tuple of int]
        Indices of the dimensions plotted on each off-diagonal subplot.
    """
    
    def __init__(
        self, 
        d=4, 
        diag=True, 
        diag_norm="max",
        diag_scale=0.65,
        diag_rspine=False,
        limits=None, 
        labels=None, 
        corner=True,
        **fig_kws
    ):
        """
        Parameters
        ----------
        d : int
            The number of rows/columns.
        diag : bool
            Whether to include diagonal subplots (univariate plots). If False,
            we have an (n - 1) x (n - 1) grid instead of an n x n grid.
        diag_norm : {"max", "area"}
            Normalize 1D histograms max value or the area under the curve.
        diag_scale : float
            Scale the 1D histograms by this value.
        diag_rspine : bool
            Whether to include right spine on diagonal subplots (if `corner`).
        limits : list[tuple], length n
            The (min, max) for each dimension. (These can be set later.)
        labels : list[str]
            The label for each dimension. (These can be set later.)
        corner : bool
            Whether to hide the upper-triangular subplots.
        **fig_kws
            Key word arguments passed to `pplt.subplots()`.
        """
        # Create figure.
        self.new = True
        self.d = self.nrows = self.ncols = d
        self.corner = corner
        self.diag = diag
        self.diag_norm = diag_norm
        self.diag_scale = diag_scale
        self.diag_rspine = diag_rspine
        if not self.diag:
            self.nrows = self.nrows - 1
            self.ncols = self.ncols - 1
        self.fig_kws = fig_kws
        self.fig_kws.setdefault("figwidth", 1.5 * self.nrows)
        self.fig_kws.setdefault("aligny", True)
        self.fig, self.axs = pplt.subplots(
            nrows=self.nrows,
            ncols=self.ncols,
            spanx=False,
            spany=False,
            sharex=False,
            sharey=False,
            **self.fig_kws,
        )
        # Collect diagonal/off-diagonal subplots and indices.
        self.diag_axs = []
        self.offdiag_axs = []
        self.offdiag_axs_u = []
        self.diag_indices = []
        self.offdiag_indices = []
        self.offdiag_indices_u = []
        if self.diag:
            for i in range(self.d):
                self.diag_axs.append(self.axs[i, i])
                self.diag_indices.append(i)
            for i in range(1, self.d):
                for j in range(i):
                    self.offdiag_axs.append(self.axs[i, j])
                    self.offdiag_axs_u.append(self.axs[j, i])
                    self.offdiag_indices.append((j, i))
                    self.offdiag_indices_u.append((i, j))
        else:
            for i in range(self.d - 1):
                for j in range(i + 1):
                    self.offdiag_axs.append(self.axs[i, j])
                    self.offdiag_indices.append((j, i + 1))

        # Set limits and labels.
        self.limits = limits
        if limits is not None:
            self.set_limits(limits)
        self.labels = labels
        if labels is not None:
            self.set_labels(labels)

        # Formatting
        if self.corner or not self.diag:
            for i in range(self.nrows):
                for j in range(self.ncols):
                    if j > i:
                        self.axs[i, j].axis("off")   
        self.axs[:-1, :].format(xticklabels=[])
        for i in range(self.nrows):
            for j in range(self.ncols):
                ax = self.axs[i, j]
                if i != self.nrows - 1:
                    ax.format(xticklabels=[])
                if j != 0:
                    if not (i == j and self.diag_rspine and self.corner and self.diag):                
                        ax.format(yticklabels=[])
        self.axs.format(xspineloc="bottom", yspineloc="left")
        if self.corner:
            if self.diag_rspine:
                self.format_diag(yspineloc="right")
            else:
                self.format_diag(yspineloc="neither")
        self.format_diag(ylim=(0.0, 1.0))
        self.axs.format(xtickminor=True, ytickminor=True, xlocator=("maxn", 3), ylocator=("maxn", 3))
            
    def format_offdiag(self, **kws):
        """Format off-diagonal subplots."""
        for ax in [self.offdiag_axs + self.offdiag_axs_u]:
            ax.format(**kws)
            
    def format_diag(self, **kws):
        """Format diagonal subplots."""
        for ax in self.diag_axs:
            ax.format(**kws)
        if not self.corner:
            for ax in self.diag_axs[1:]:
                ax.format(yticklabels=[])
        self._fake_diag_yticks()

    def get_labels(self):
        """Return the dimension labels."""
        if self.diag:
            labels = [ax.get_xlabel() for ax in self.diag_axs]
        else:
            labels = [self.axs[-1, i].get_xlabel() for i in range(self.d - 1)]
            labels = labels + [self.axs[-1, 0].get_ylabel()]
        return labels

    def set_labels(self, labels):
        """Set the dimension labels."""
        for ax, label in zip(self.axs[-1, :], labels):
            ax.format(xlabel=label)
        for ax, label in zip(self.axs[int(self.diag):, 0], labels[1:]):
            ax.format(ylabel=label)
        if self.diag and not self.corner:
            self.axs[0, 0].format(ylabel=labels[0])
        self.labels = labels

    def get_limits(self):
        """Return the plot limits."""
        if self.diag:
            limits = [ax.get_xlim() for ax in self.diag_axs]
        else:
            limits = [self.axs[-1, i].get_xlim() for i in range(self.d - 1)]
            limits = limits + [self.axs[-1, 0].get_ylim()]
        return limits

    def set_limits(self, limits=None, expand=False):
        """Set the plot limits.

        Parameters
        ----------
        limits : list[tuple], length n
            The (min, max) for each dimension.
        expand : bool
            If True, compare the proposed limits to the existing limits, expanding
            if the new limits are wider.
        """
        if limits is not None:
            if expand:
                limits = np.array(limits)
                limits_old = np.array(self.get_limits())
                mins = np.minimum(limits[:, 0], limits_old[:, 0])
                maxs = np.maximum(limits[:, 1], limits_old[:, 1])
                limits = list(zip(mins, maxs))
            for (i, j), ax in zip(self.offdiag_indices, self.offdiag_axs):
                ax.format(ylim=limits[j])
            for i in range(self.ncols):
                self.axs[:, i].format(xlim=limits[i])
        self.limits = self.get_limits()
        
    def _fake_diag_yticks(self):
        """The yticks on the (0, 0) subplot correspond to the other subplots in the row.
        
        Source: pandas.plotting.scatterplot_matrix.
        """
        if self.corner or not self.diag:
            return
        limits = self.limits
        if limits is None:
            limits = self.get_limits()
            
        lim1 = limits[0]
        locs = self.axs[0, 0].xaxis.get_majorticklocs()
        locs = locs[(lim1[0] <= locs) & (locs <= lim1[1])]
        adj = (locs - lim1[0]) / (lim1[1] - lim1[0])

        lim0 = self.axs[0, 0].get_ylim()
        adj = adj * (lim0[1] - lim0[0]) + lim0[0]
        self.axs[0, 0].yaxis.set_ticks(adj)

        if np.all(locs == locs.astype(int)):
            locs = locs.astype(int)
        self.axs[0, 0].yaxis.set_ticklabels(locs)
        
    def plot_diag(self, x, y, axis=0, **kws):
        """Plot data on diagonal subplot. The plot is assumed to represent
        a probability density, so is normalized by area."""
        if self.diag_norm == "max":
            y = y / np.max(y)
        elif self.diag_norm == "area":
            dx = np.abs(np.diff(x)[0])
            y = y / (np.sum(y) * dx)
        y = y * self.diag_scale
        return plot1d(x, y, ax=self.diag_axs[axis], **kws)

    def plot_image(
        self,
        f,
        coords=None,
        prof_edge_only=False,
        lower=True,
        upper=True,
        diag=True,
        update_limits=True,
        diag_kws=None,
        **kws,
    ):
        """Plot an image.

        Parameters
        ----------
        f : ndarray
            A d-dimensional image.
        coords : list[ndarray]
            Coordinates along each axis of the grid (if `data` is an image).
        prof_edge_only : bool
            If plotting profiles on top of images (on off-diagonal subplots), whether
            to plot x profiles only in bottom row and y profiles only in left column.
        lower, upper, diag : bool
            Whether to plot on the lower triangular, upper triangular, and/or diagonal subplots.
        update_limits : bool
            Whether to extend the existing plot limits.
        diag_kws : dict
            Key word argument passed to `visualization.plot1d`.
        **kws
            Key word arguments pass to `visualization.image.plot2d`
        """
        if diag_kws is None:
            diag_kws = dict()
        diag_kws.setdefault("color", "black")
        diag_kws.setdefault("lw", 1.0)
        diag_kws.setdefault("kind", "step")
        kws.setdefault("kind", "pcolor")
        kws.setdefault("profx", False)
        kws.setdefault("profy", False)

        if coords is None:
            coords = [np.arange(f.shape[i]) for i in range(f.ndim)]

        if update_limits:
            edges = [psdist.utils.edges_from_centers(c) for c in coords]
            limits = [(np.min(e), np.max(e)) for e in edges]
            self.set_limits(limits, expand=(not self.new))
        self.new = False

        # Univariate plots.
        if diag:
            for ax, axis in zip(self.diag_axs, self.diag_indices):
                profile = psdist.image.project(f, axis=axis)
                self.plot_diag(coords[axis], profile, axis=axis, **diag_kws)

        # Bivariate plots.
        profx, profy = [kws.pop(key) for key in ("profx", "profy")]
        if lower:
            for ax, axis in zip(self.offdiag_axs, self.offdiag_indices):
                if prof_edge_only:
                    if profx:
                        kws["profx"] = axis[1] == self.d - 1
                    if profy:
                        kws["profy"] = axis[0] == 0
                _f = psdist.image.project(f, axis=axis)
                _f = _f / np.max(_f)
                _coords = [coords[k] for k in axis]
                vis_image.plot2d(_f, coords=_coords, ax=ax, **kws)
        if not self.corner and upper:
            for ax, axis in zip(self.offdiag_axs_u, self.offdiag_indices_u):
                _f = psdist.image.project(f, axis=axis)
                _f = _f / np.max(_f)
                _coords = [coords[k] for k in axis]
                vis_image.plot2d(_f, coords=_coords, ax=ax, **kws)
        self._fake_diag_yticks()
            
    def plot_cloud(
        self,
        X,
        limits=None,
        bins="auto",
        autolim_kws=None,
        prof_edge_only=False,
        lower=True,
        upper=True,
        diag=True,
        update_limits=True,
        diag_kws=None,
        **kws,
    ):
        """Plot a point cloud.

        Parameters
        ----------
        X : ndarray, shape (n, d)
            Coordinates of n points in d-dimensional space.
        limits : list[tuple], length d
            The (min, max) axis limits.
        bins : 'auto', int, list[int]
            The number of bins along each dimension (if plot type requires histogram
            computation). If int or 'auto', applies to all dimensions. Currently
            the histogram is computed with limits based on the data min/max, not
            the plot limits.
        prof_edge_only : bool
            If plotting profiles on top of images (on off-diagonal subplots), whether
            to plot x profiles only in bottom row and y profiles only in left column.
        lower, upper, diag : bool
            Whether to plot on the lower triangular, upper triangular, and/or diagonal subplots.
        update_limits : bool
            Whether to extend the existing plot limits.
        diag_kws : dict
            Key word argument passed to `visualization.plot1d`.
        **kws
            Key word arguments pass to `visualization.cloud.plot2d`
        """
        if diag_kws is None:
            diag_kws = dict()
        diag_kws.setdefault("color", "black")
        diag_kws.setdefault("lw", 1.0)
        diag_kws.setdefault("kind", "step")
        kws.setdefault("kind", "hist")
        kws.setdefault("profx", False)
        kws.setdefault("profy", False)

        if limits is None:
            if autolim_kws is None:
                autolim_kws = dict()
            autolim_kws.setdefault("pad", 0.1)
            limits = vis_cloud.auto_limits(X, **autolim_kws)
        if update_limits:
            self.set_limits(limits, expand=(not self.new))
        limits = self.get_limits()
        self.new = False
        
        if not psdist.utils.array_like(bins):
            bins = X.shape[1] * [bins]

        # Univariate plots. Remember histogram bins and use them for 2D histograms.
        edges = []
        for axis in range(self.d):
            if psdist.utils.array_like(bins[axis]):
                _edges = bins[axis]
            else:
                _edges = np.histogram_bin_edges(X[:, axis], bins[axis], limits[axis])
            edges.append(_edges)
            if self.diag and diag:
                heights, _ = np.histogram(X[:, axis], _edges)
                centers = psdist.utils.centers_from_edges(_edges)
                self.plot_diag(centers, heights, axis=axis, **diag_kws)

        # Bivariate plots:
        profx, profy = [kws.pop(key) for key in ("profx", "profy")]
        if lower:
            for ax, axis in zip(self.offdiag_axs, self.offdiag_indices):
                if prof_edge_only:
                    if profx:
                        kws["profx"] = axis[1] == self.d - 1
                    if profy:
                        kws["profy"] = axis[0] == 0
                if kws["kind"] in ["hist", "contour", "contourf"]:
                    kws["bins"] = [edges[axis[0]], edges[axis[1]]]
                vis_cloud.plot2d(X[:, axis], ax=ax, **kws)
        if upper and not self.corner:
            for ax, axis in zip(self.offdiag_axs_u, self.offdiag_indices_u):
                if kws["kind"] in ["hist", "contour", "contourf"]:
                    kws["bins"] = [edges[axis[0]], edges[axis[1]]]
                vis_cloud.plot2d(X[:, axis], ax=ax, **kws)
        self._fake_diag_yticks()


class SliceGrid:
    """Grid for slice matrix plots (https://arxiv.org/abs/2301.04178).

    This plot is used to visualize four dimensions of a distribution f(x1, x2, x3, x4).

    The main panel is an nrows x ncols grid that shows f(x1, x2 | x3, x4) -- the
    x1-x2 distribution for a planar slice in x3-x4. Each subplot corresponds to a
    different location in the x3-x4 plane.

    The following is only included if `marginals` is True:

        The bottom panel shows the marginal 3D distribution f(x1, x2 | x3).

        The right panel shows the marginal 3D distribution f(x1, x2 | x4).

        The bottom right subplot shows the full projection f(x1, x2).

        The lone subplot on the bottom right shows f(x1, x2)l, the full projection
        onto the x1-x2 plane.

    Attributes
    ----------
    fig : proplot.figure.Figure
        The main figure.
    axs : proplot.gridspec.SubplotGrid
        The subplot axes.
    _axs : proplot.figure.Figure
        The subplot axes on the main panel.
    _axs_marg_x, _axs_marg_y, _axs_marg_xy : proplot.gridspec.SubplotGrid
        The subplot axes on the marginal panels.
    """

    def __init__(
        self,
        nrows=9,
        ncols=9,
        space=0.0,
        gap=2.0,
        marginals=True,
        annotate=True,
        annotate_kws_view=None,
        annotate_kws_slice=None,
        slice_label_height=0.22,
        **fig_kws,
    ):
        """Constructor.

        nrows, ncols : int
            The number of rows/colums in the figure.
        space : float
            Spacing between subplots.
        gap : float
            Gap between main and marginal panels.
        marginals : bool
            Whether to include the marginal panels. If they are not included, we just
            have an nrows x ncols grid.
        annotate : bool
            Whether to add dimension labels/arrows to the figure.
        annotate_kws_view, annotate_kws_slice : dict
            Key word arguments for figure text. The 'view' key words are for the view
            dimension labels; they are printed on top of one of the subplots. The
            'slice' key words are for the slice dimension labels; they are printed
            on the sides of the figure, between the main and marginal panels.
        slice_label_height : float
            Tweaks the position of slice labels. Need a better way to handle this.
        **fig_kws
            Key word arguments for `pplt.subplots`.
        """
        self.nrows = nrows
        self.ncols = ncols
        self.space = space
        self.gap = gap
        self.marginals = marginals
        self.annotate = annotate
        self.slice_label_height = slice_label_height
        self.fig_kws = fig_kws
        self.axis_slice = None
        self.axis_view = None
        self.ind_slice = None

        self.annotate_kws_view = annotate_kws_view
        if self.annotate_kws_view is None:
            self.annotate_kws_view = dict()
        self.annotate_kws_view.setdefault("color", "black")
        self.annotate_kws_view.setdefault("xycoords", "axes fraction")
        self.annotate_kws_view.setdefault("horizontalalignment", "center")
        self.annotate_kws_view.setdefault("verticalalignment", "center")

        self.annotate_kws_slice = annotate_kws_slice
        if self.annotate_kws_slice is None:
            self.annotate_kws_slice = dict()
        self.annotate_kws_slice.setdefault("color", "black")
        self.annotate_kws_slice.setdefault("xycoords", "axes fraction")
        self.annotate_kws_slice.setdefault("horizontalalignment", "center")
        self.annotate_kws_slice.setdefault("verticalalignment", "center")
        self.annotate_kws_slice.setdefault(
            "arrowprops", dict(arrowstyle="->", color="black")
        )

        fig_kws.setdefault("figwidth", 8.5 * (ncols / 13.0))
        fig_kws.setdefault("share", True)
        fig_kws["ncols"] = ncols + 1 if marginals else ncols
        fig_kws["nrows"] = nrows + 1 if marginals else nrows
        hspace = nrows * [space]
        wspace = ncols * [space]
        if marginals:
            hspace[-1] = wspace[-1] = gap
        else:
            hspace = hspace[:-1]
            wspace = wspace[:-1]
        fig_kws["hspace"] = hspace
        fig_kws["wspace"] = wspace

        self.fig, self.axs = pplt.subplots(**fig_kws)

        self._axs = self.axs[:-1, :-1]
        self._axs_marg_x = []
        self._axs_marg_y = []
        if self.marginals:
            self._axs_marg_x = self.axs[-1, :]
            self._axs_marg_y = self.axs[:, -1]
        self._ax_marg_xy = self.axs[-1, -1]

    def _annotate(
        self,
        labels=None,
        slice_label_height=0.22,
        annotate_kws_view=None,
        annotate_kws_slice=None,
    ):
        """Add dimension labels and arrows."""
        # Label the view dimensions.
        for i, xy in enumerate([(0.5, 0.13), (0.12, 0.5)]):
            self.axs[0, 0].annotate(labels[i], xy=xy, **self.annotate_kws_view)

        # Label the slice dimensions. Print dimension labels with arrows like this:
        # "<----- x ----->" on the bottom and right side of the main panel.
        arrow_length = 2.5  # arrow length
        text_length = 0.15  # controls space between dimension label and start of arrow
        i = -1 - int(self.marginals)
        anchors = (self.axs[i, self.ncols // 2], self.axs[self.nrows // 2, i])
        anchors[0].annotate(
            labels[2], xy=(0.5, -slice_label_height), **annotate_kws_slice
        )
        anchors[1].annotate(
            labels[3], xy=(1.0 + slice_label_height, 0.5), **annotate_kws_slice
        )
        for arrow_direction in (1.0, -1.0):
            anchors[0].annotate(
                "",
                xy=(0.5 + arrow_direction * arrow_length, -slice_label_height),
                xytext=(0.5 + arrow_direction * text_length, -slice_label_height),
                **annotate_kws_slice,
            )
            anchors[1].annotate(
                "",
                xy=(1.0 + slice_label_height, 0.5 + arrow_direction * arrow_length),
                xytext=(1.0 + slice_label_height, 0.5 + arrow_direction * text_length),
                **annotate_kws_slice,
            )

    def get_ind_slice(self):
        """Return slice indices from latest plot call."""
        return self.ind_slice

    def get_axis_slice(self):
        """Return slice axis from latest plot call."""
        return self.axis_slice

    def get_axis_view(self):
        """Return view axis from latest plot call."""
        return self.axis_view

    def set_limits(self, limits):
        """Set the plot limits."""
        for ax in self.axs:
            ax.format(xlim=limits[0], ylim=limits[1])

    def plot_image(
        self,
        f,
        coords=None,
        labels=None,
        axis_view=(0, 1),
        axis_slice=(2, 3),
        pad=0.0,
        debug=False,
        **kws,
    ):
        """Plot a d-dimensional image.

        Parameters
        ----------
        f : ndarray
            An d-dimensional image.
        coords : list[ndarray]
            Coordinates along each axis of the grid (if `data` is an image).
        labels : list[str], length n
            Label for each dimension.
        axis_view, axis_slice : 2-tuple of int
            The axis to view (plot) and to slice.
        pad : int, float, list
            This determines the start/stop indices along the sliced dimensions. If
            0, space the indices along axis `i` uniformly between 0 and `f.shape[i]`.
            Otherwise, add a padding equal to `int(pad[i] * f.shape[i])`. So, if
            the shape=10 and pad=0.1, we would start from 1 and end at 9.
        debug : bool
            Whether to print debugging messages.
        **kws
            Key word arguments pass to `visualization.image.plot2d`
        """
        # Setup
        # -----------------------------------------------------------------------
        if f.ndim < 4:
            raise ValueError(f"f.ndim = {f.ndim} < 4")
        if coords is None:
            coords = [np.arange(s) for s in f.shape]
        self.axis_view = axis_view
        self.axis_slice = axis_slice

        # Compute 4D/3D/2D projections.
        _f = psdist.image.project(f, axis_view + axis_slice)
        _fx = psdist.image.project(f, axis_view + axis_slice[:1])
        _fy = psdist.image.project(f, axis_view + axis_slice[1:])
        _fxy = psdist.image.project(f, axis_view)

        # Compute new coords and labels.
        _coords = [coords[i] for i in axis_view + axis_slice]
        _labels = None
        if labels is not None:
            _labels = [labels[i] for i in axis_view + axis_slice]

        # Select slice indices.
        if type(pad) in [float, int]:
            pad = len(axis_slice) * [pad]
        ind_slice = []
        for i, nsteps, _pad in zip(axis_slice, [self.nrows, self.ncols], pad):
            start = int(_pad * f.shape[i])
            stop = f.shape[i] - 1 - start
            if stop - start == nsteps - 1:
                ind_slice.append(np.arange(nsteps))
            elif (stop - start) < nsteps:
                raise ValueError(f"f.shape[{i}] < number of slice indices requested.")
            ind_slice.append(np.linspace(start, stop, nsteps).astype(int))
        ind_slice = tuple(ind_slice)
        self.ind_slice = ind_slice

        if debug:
            print("Slice indices:")
            for ind in ind_slice:
                print(ind)

        # Slice the 4D projection. The axes order was already handled by `project`;
        # the first two axes are the view axes and the last two axes are the
        # slice axes.
        axis_view = (0, 1)
        axis_slice = (2, 3)
        idx = 4 * [slice(None)]
        for axis, ind in zip(axis_slice, ind_slice):
            idx[axis] = ind
            _f = _f[tuple(idx)]
            idx[axis] = slice(None)

        # Slice the 3D projections.
        _fx = _fx[:, :, ind_slice[0]]
        _fy = _fy[:, :, ind_slice[1]]

        # Slice coords.
        for i, ind in zip(axis_slice, ind_slice):
            _coords[i] = _coords[i][ind]

        # Normalize each distribution.
        _f = psdist.image.process(_f, norm="max")
        _fx = psdist.image.process(_fx, norm="max")
        _fy = psdist.image.process(_fy, norm="max")
        _fxy = psdist.image.process(_fxy, norm="max")

        if debug:
            print("_f.shape =", _f.shape)
            print("_fx.shape =", _fx.shape)
            print("_fy.shape =", _fy.shape)
            print("_fxy.shape =", _fxy.shape)
            for i in range(_f.ndim):
                assert _f.shape[i] == len(_coords[i])

        # Add dimension labels to the figure.
        if self.annotate and _labels is not None:
            self._annotate(
                labels=_labels,
                slice_label_height=self.slice_label_height,
                annotate_kws_view=self.annotate_kws_view,
                annotate_kws_slice=self.annotate_kws_slice,
            )

        # Plotting
        # -----------------------------------------------------------------------
        for i in range(self.nrows):
            for j in range(self.ncols):
                ax = self.axs[self.nrows - 1 - i, j]
                idx = psdist.image.slice_idx(
                    _f.ndim, axis=axis_slice, ind=[(j, j + 1), (i, i + 1)]
                )
                vis_image.plot2d(
                    psdist.image.project(_f[idx], axis_view),
                    coords=[_coords[axis_view[0]], _coords[axis_view[1]]],
                    ax=ax,
                    **kws,
                )
        if self.marginals:
            for i, ax in enumerate(reversed(self.axs[:-1, -1])):
                vis_image.plot2d(
                    _fy[:, :, i],
                    coords=[_coords[axis_view[0]], _coords[axis_view[1]]],
                    ax=ax,
                    **kws,
                )
            for i, ax in enumerate(self.axs[-1, :-1]):
                vis_image.plot2d(
                    _fx[:, :, i],
                    [_coords[axis_view[0]], _coords[axis_view[1]]],
                    ax=ax,
                    **kws,
                )
            vis_image.plot2d(
                _fxy,
                coords=[_coords[axis_view[0]], _coords[axis_view[1]]],
                ax=self.axs[-1, -1],
                **kws,
            )

    def _plot_cloud(
        self,
        X,
        labels=None,
        bins_slice="auto",
        axis_view=(0, 1),
        axis_slice=(2, 3),
        pad=0.0,
        debug=False,
        autolim_kws=None,
        **kws,
    ):
        """Plot a point cloud.

        NOTE: this is not currently working... for the time being, it is recommended
        to generate a 4D histogram and then call `plot_image`.

        Parameters
        ----------
        X : ndarray, shape (n, d)
            Coordinates of n points in d-dimensional space.
        labels : list[str], length d
            Label for each dimension.
        bins_slice : int, list[int], list[ndarray]
            Specifies the bins used for slicing in `axis_slice`. The bin range
            is determined by the min/max point in `X`.
        axis_view, axis_slice : 2-tuple of int
            The axis to view (plot) and to slice.
        pad : int, float, list
            Fractional padding added to the start/stop indices for the bins in the sliced
            dimensions.
        debug : bool
            Whether to print debugging messages.
        **kws
            Key word arguments pass to `visualization.cloud.plot2d`
        """

        warnings.warn(
            "This is not currently working. For the time being, it is recommended to generate a 4D histogram and call `plot_image`."
        )

        # Setup
        # -----------------------------------------------------------------------
        if X.shape[1] < 4:
            raise ValueError(f"X.shape[1] = {X.shape[1]} < 4")
        self.axis_view = axis_view
        self.axis_slice = axis_slice

        edges_slice = psdist.cloud.histogram_bin_edges(X[:, axis_slice], bins_slice)

        # Select slice indices.
        if type(pad) in [float, int]:
            pad = len(axis_slice) * [pad]
        ind_slice = []
        for i, (steps, _pad) in enumerate(zip([self.nrows, self.ncols], pad)):
            start = int(_pad * len(edges_slice[i]))
            stop = len(edges_slice[i]) - 1 - start
            if (stop - start) < steps:
                raise ValueError(f"Too many slices requested.")
            ind_slice.append(np.linspace(start, stop, steps + 1).astype(int))
        ind_slice = tuple(ind_slice)
        self.ind_slice = ind_slice

        # Slice the bin indices.
        edges_slice = [e[ind] for e, ind in zip(edges_slice, ind_slice)]

        # Add dimension labels to the figure.
        if labels is not None:
            if self.annotate and labels is not None:
                self._annotate(
                    labels=[labels[k] for k in axis_view + axis_slice],
                    slice_label_height=self.slice_label_height,
                    annotate_kws_view=self.annotate_kws_view,
                    annotate_kws_slice=self.annotate_kws_slice,
                )

        # Plotting
        # -----------------------------------------------------------------------
        for i in range(self.nrows):
            for j in range(self.ncols):
                ax = self.axs[self.nrows - 1 - i, j]
                _X = psdist.cloud.slice_planar(
                    X,
                    axis=axis_slice,
                    center=[
                        0.5 * (edges_slice[0][j] + edges_slice[0][j + 1]),
                        0.5 * (edges_slice[1][i] + edges_slice[1][i + 1]),
                    ],
                    width=[
                        np.abs(edges_slice[0][j] - edges_slice[1][j + 1]),
                        np.abs(edges_slice[1][i] - edges_slice[1][i + 1]),
                    ],
                )
                vis_cloud.plot2d(_X[:, axis_slice], ax=ax, **kws)



"""Plotting routines for multi-dimensional images."""
from ipywidgets import interactive
from ipywidgets import widgets
from matplotlib import pyplot as plt
import numpy as np
import proplot as pplt

import psdist.image
from psdist.utils import edges_from_centers
import psdist.visualization as vis


def plot_profiles(
    f,
    coords=None,
    ax=None,
    profx=True,
    profy=True,
    scale=0.12,
    start="edge",
    **kws,
):
    """Plot the one-dimensional profiles of a two-dimensional image.

    Parameters
    ----------
    f : ndarray
        A two-dimensional image.
    coords: (xcoords, ycoords)
        Lists specifying pixel center coordinates along each axis.
    ax : matplotlib.pyplt.Axes
        The axis on which to plot.
    profx, profy : bool
        Whether to plot the x/y profile.
    scale : float
        Maximum of the 1D plot relative to the axes limits.
    start : {'edge', 'center'}
        Whether to start the plot at the center or edge of the first row/column.
    **kws
        Key word arguments passed to `psdist.visualization.plot1d`.
    """
    kws.setdefault("kind", "step")
    kws.setdefault("lw", 1.0)
    kws.setdefault("color", "white")
    kws.setdefault("alpha", 0.6)

    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]

    edges = [edges_from_centers(c) for c in coords]
    for axis, proceed in enumerate([profx, profy]):
        if proceed:
            profile = psdist.image.project(f, axis=axis)
            # Scale
            profile_max = np.max(profile)
            if profile_max > 0.0:
                profile = profile / profile_max
            j = int(axis == 0)
            profile = profile * scale * np.abs(coords[j][-1] - coords[j][0])
            # Start from ax spine.
            offset = 0.0
            if start == "edge":
                offset = edges[j][0]
            elif start == "center":
                offset = coords[j][0]
            vis.plot1d(coords[axis], profile, ax=ax, offset=offset, flipxy=axis, **kws)
    return ax


def plot_rms_ellipse(
    f, coords=None, ax=None, level=1.0, center_at_mean=True, **ellipse_kws
):
    """Compute and plot the RMS ellipse from a two-dimensional image.

    Parameters
    ----------
    f : ndarray
        A two-dimensional image.
    coords: (xcoords, ycoords)
        Lists specifying pixel center coordinates along each axis.
    ax : matplotlib.pyplt.Axes
        The axis on which to plot.
    level : number of list of numbers
        If a number, plot the rms ellipse inflated by the number. If a list
        of numbers, repeat for each number.
    center_at_mean : bool
        Whether to center the ellipse at the image centroid.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]
    center = psdist.image.mean(f, coords) if center_at_mean else (0.0, 0.0)
    Sigma = psdist.image.cov(f, coords)
    return vis.rms_ellipse(Sigma, center, level=level, ax=ax, **ellipse_kws)


def plot2d(
    f,
    coords=None,
    ax=None,
    kind="pcolor",
    profx=False,
    profy=False,
    prof_kws=None,
    process_kws=None,
    offset=None,
    offset_type="relative",
    mask=False,
    rms_ellipse=False,
    rms_ellipse_kws=None,
    return_mesh=False,
    **kws,
):
    """Plot a two-dimensional image.

    Parameters
    ----------
    f : ndarray
        A two-dimensional image.
    coords: (xcoords, ycoords)
        Lists specifying pixel center coordinates along each axis.
    ax : matplotlib.pyplt.Axes
        The axis on which to plot.
    kind : ['pcolor', 'contour', 'contourf']
        Whether to call `ax.pcolormesh`, `ax.contour`, or `ax.contourf`.
    profx, profy : bool
        Whether to plot the x/y profile.
    prof_kws : dict
        Key words arguments for `image_profiles`.
    rms_ellipse : bool
        Whether to plot rms ellipse.
    rms_ellipse_kws : dict
        Key word arguments for `image_rms_ellipse`.
    return_mesh : bool
        Whether to return a mesh from `ax.pcolormesh`.
    process_kws : dict
        Key word arguments passed to `psdist.image.process`.
    offset, offset_type : float, {"relative", "absolute"}
        Adds offset to the image (helpful to get rid of zeros for logarithmic 
        color scales. If offset_type is 'relative' add `min(f) * offset` to
        the image. Otherwise add `offset`.
    mask : bool
        Whether to plot pixels at or below zero.
    **kws
        Key word arguments passed to plotting function.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]

    # Process key word arguments.
    if process_kws is None:
        process_kws = dict()

    if rms_ellipse_kws is None:
        rms_ellipse_kws = dict()

    func = None
    if kind == "pcolor":
        func = ax.pcolormesh
        kws.setdefault("ec", "None")
        kws.setdefault("linewidth", 0.0)
        kws.setdefault("rasterized", True)
        kws.setdefault("shading", "auto")
    elif kind == "contour":
        func = ax.contour
    elif kind == "contourf":
        func = ax.contourf
    else:
        raise ValueError("Invalid plot kind.")
    kws.setdefault("colorbar", False)
    kws.setdefault("colorbar_kw", dict())

    # Process the image.
    f = f.copy()
    f = psdist.image.process(f, **process_kws)
    if offset is not None:
        if offset_type == "relative" and np.count_nonzero(f):
            offset = offset * np.min(f[f > 0])
        f = f + offset
    else:
        offset = 0.0

    # Make sure there are no more zero elements if norm='log'.
    log = "norm" in kws and kws["norm"] == "log"
    if log:
        kws["colorbar_kw"]["formatter"] = "log"
    if mask or log:
        f = np.ma.masked_less_equal(f, 0)

    # If there are only zero elements, increase vmax so that the
    # lowest color is plotted.
    if not np.count_nonzero(f):
        kws["vmin"] = 1.0
        kws["vmax"] = 1.0

    # Plot.
    mesh = func(coords[0].T, coords[1].T, f.T, **kws)
    if rms_ellipse:
        if rms_ellipse_kws is None:
            rms_ellipse_kws = dict()
        plot_rms_ellipse(f, coords=coords, ax=ax, **rms_ellipse_kws)
    if profx or profy:
        if prof_kws is None:
            prof_kws = dict()
        if kind == "contourf":
            prof_kws.setdefault("start", "center")
        plot_profiles(f - offset, coords=coords, ax=ax, profx=profx, profy=profy, **prof_kws)
    if return_mesh:
        return ax, mesh
    else:
        return ax


def joint(f, coords=None, grid_kws=None, marg_kws=None, **kws):
    """Joint plot.

    This is a convenience function; see `psdist.visualization.grid.JointGrid`.

    Parameters
    ----------
    f : ndarray
        A 2-dimensional image.
    coords : list[ndarray]
        Coordinates along each dimension of `f`.
    grid_kws : dict
        Key word arguments passed to `JointGrid`.
    marg_kws : dict
        Key word arguments passed to `visualization.plot1d`.
    **kws
        Key word arguments passed to `visualization.image.plot2d.`

    Returns
    -------
    psdist.visualization.grid.JointGrid
    """
    from psdist.visualization.grid import JointGrid
    if grid_kws is None:
        grid_kws = dict()
    grid = JointGrid(**grid_kws)
    grid.plot_image(f, coords, marg_kws=marg_kws, **kws)
    return grid


def corner(
    f,
    coords=None,
    labels=None,
    prof_edge_only=False,
    update_limits=True,
    diag_kws=None,
    grid_kws=None,
    **kws,
):
    """Corner plot (scatter plot matrix).

    This is a convenience function; see `psdist.visualization.grid.CornerGrid`.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinates along each dimension of `f`.
    labels : list[str], length n
        Label for each dimension.
    axis_view, axis_slice : 2-tuple of int
        The axis to view (plot) and to slice.
    pad : int, float, list
        This determines the start/stop indices along the sliced dimensions. If
        0, space the indices along axis `i` uniformly between 0 and `f.shape[i]`.
        Otherwise, add a padding equal to `int(pad[i] * f.shape[i])`. So, if
        the shape=10 and pad=0.1, we would start from 1 and end at 9.
    debug : bool
        Whether to print debugging messages.
    **kws
        Key word arguments pass to `visualization.image.plot2d`

    Returns
    -------
    psdist.visualization.grid.CornerGrid
        The `CornerGrid` on which the plot was drawn.
    """
    from psdist.visualization.grid import CornerGrid
    if grid_kws is None:
        grid_kws = dict()
    grid = CornerGrid(d=f.ndim, **grid_kws)
    if labels is not None:
        grid.set_labels(labels)
    grid.plot_image(
        f,
        coords=coords,
        prof_edge_only=prof_edge_only,
        update_limits=True,
        diag_kws=diag_kws,
        **kws,
    )
    return grid


def slice_matrix(
    f,
    coords=None,
    labels=None,
    axis_view=(0, 1),
    axis_slice=(2, 3),
    pad=0.0,
    debug=False,
    grid_kws=None,
    **kws,
):
    """Slice matrix plot.

    This is a convenience function; see `psdist.visualization.grid.SliceGrid`.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinates along each axis of the grid (if `data` is an image).
    labels : list[str], length n
        Label for each dimension.
    axis_view, axis_slice : 2-tuple of int
        The axis to view (plot) and to slice.
    pad : int, float, list
        This determines the start/stop indices along the sliced dimensions. If
        0, space the indices along axis `i` uniformly between 0 and `f.shape[i]`.
        Otherwise, add a padding equal to `int(pad[i] * f.shape[i])`. So, if
        the shape=10 and pad=0.1, we would start from 1 and end at 9.
    debug : bool
        Whether to print debugging messages.
    grid_kws : dict
        Key word arguments passed to `visualization.grid.SliceGrid`.
    **kws
        Key word arguments passed to `visualization.image.plot2d`

    Returns
    -------
    psdist.visualization.grid.SliceGrid
        The `SliceGrid` on which the plot was drawn.
    """
    from psdist.visualization.grid import SliceGrid
    if grid_kws is None:
        grid_kws = dict()
    grid_kws.setdefault("space", 0.2)
    grid_kws.setdefault("annotate_kws_view", dict(color="white"))
    grid_kws.setdefault("annotate_kws_slice", dict(color="black"))
    grid_kws.setdefault("xticks", [])
    grid_kws.setdefault("yticks", [])
    grid_kws.setdefault("xspineloc", "neither")
    grid_kws.setdefault("yspineloc", "neither")
    grid = SliceGrid(**grid_kws)
    grid.plot_image(
        f,
        coords=None,
        labels=None,
        axis_view=(0, 1),
        axis_slice=(2, 3),
        pad=0.0,
        debug=False,
    )
    return grid


def proj2d_interactive_slice(
    f,
    coords=None,
    default_ind=(0, 1),
    slice_type="int",
    dims=None,
    units=None,
    cmaps=None,
    thresh_slider=False,
    profiles_checkbox=False,
    **plot_kws,
):
    """2D partial projection with interactive slicing.

    The distribution is projected onto the specified axes. Sliders provide the
    option to slice the distribution before projecting.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Coordinate arrays along each dimension. A square grid is assumed.
    default_ind : (i, j)
        Default x and y index to plot.
    slice_type : {'int', 'range'}
        Whether to slice one index along the axis or a range of indices.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    cmaps : list
        Color map options for dropdown menu.
    thresh_slider : bool
        Whether to include a threshold slider.
    profiles_checkbox : bool
        Whether to include a profiles checkbox.
    **plot_kws
        Key word arguments for `visualization.image.plot2d`.

    Returns
    -------
    ipywidgets.widgets.interaction.interactive
        This widget can be displayed by calling `IPython.display.display(gui)`.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]
    if dims is None:
        dims = [f"x{i + 1}" for i in range(f.ndim)]
    if units is None:
        units = f.ndim * [""]
    dims_units = []
    for dim, unit in zip(dims, units):
        dims_units.append(f"{dim}" + f" [{unit}]" if unit != "" else dim)
    plot_kws.setdefault("colorbar", True)
    plot_kws["process_kws"] = dict(thresh_type="frac")

    # Widgets
    cmap = None
    if cmaps is not None:
        cmap = widgets.Dropdown(options=cmaps, description="cmap")
    if thresh_slider:
        thresh_slider = widgets.FloatSlider(
            value=-3.3,
            min=-8.0,
            max=0.0,
            step=0.1,
            description="thresh (log)",
            continuous_update=True,
        )
    discrete = widgets.Checkbox(value=False, description="discrete")
    log = widgets.Checkbox(value=False, description="log")
    if profiles_checkbox:
        _profiles_checkbox = widgets.Checkbox(value=False, description="profiles")
    else:
        _profiles_checkbox = None
    dim1 = widgets.Dropdown(options=dims, index=default_ind[0], description="dim 1")
    dim2 = widgets.Dropdown(options=dims, index=default_ind[1], description="dim 2")

    # Sliders and checkboxes (for slicing). Each unplotted dimension has a
    # checkbox which determine if that dimension is sliced. The slice
    # indices are determined by the slider.
    sliders, checks = [], []
    for k in range(f.ndim):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=f.shape[k],
                value=(f.shape[k] // 2),
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                value=(0, f.shape[k]),
                min=0,
                max=f.shape[k],
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("`slice_type` must be 'int' or 'range'.")
        slider.layout.display = "none"
        sliders.append(slider)
        checks.append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide/show sliders."""
        for k in range(f.ndim):
            # Hide elements for dimensions being plotted.
            valid = dims[k] not in (dim1.value, dim2.value)
            disp = None if valid else "none"
            for element in [sliders[k], checks[k]]:
                element.layout.display = disp
            # Uncheck boxes for dimensions being plotted.
            if not valid and checks[k].value:
                checks[k].value = False
            # Make sliders respond to check boxes.
            if not checks[k].value:
                sliders[k].layout.display = "none"

    # Update the slider list automatically.
    for element in (dim1, dim2, *checks):
        element.observe(hide, names="value")
    # Initial hide
    for k in range(f.ndim):
        if k in default_ind:
            checks[k].layout.display = "none"
            sliders[k].layout.display = "none"

    def update(**kws):
        """Update the figure."""
        dim1, dim2 = kws["dim1"], kws["dim2"]
        ind, checks = [], []
        for i in range(100):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                ind.append(kws[f"slider{i}"])
                
        # Return nothing if input does not make sense.
        for dim, check in zip(dims, checks):
            if check and dim in (dim1, dim2):
                return
        if dim1 == dim2:
            return
            
        # Slice and project the distribution.
        axis_view = [dims.index(dim) for dim in (dim1, dim2)]
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        for k in range(f.ndim):
            if type(ind[k]) is int:
                ind[k] = (ind[k], ind[k] + 1)
        ind = [ind[k] for k in axis_slice]
        idx = psdist.image.slice_idx(f.ndim, axis_slice, ind)
        _f = psdist.image.project(f[idx], axis_view)
        
        # Update plotting key word arguments.
        if "cmap" in kws:
            plot_kws["cmap"] = kws["cmap"]
        if "profiles_checkbox" in kws:
            plot_kws["profx"] = plot_kws["profy"] = kws["profiles_checkbox"]
        plot_kws["norm"] = "log" if kws["log"] else None
        plot_kws["process_kws"]["fill_value"] = 0
        if "thresh_slider" in kws:
            plot_kws["process_kws"]["thresh"] = 10.0 ** kws["thresh_slider"]
        else:
            plot_kws["process_kws"]["thresh"] = None
            
        # Plot the projection onto the specified axes.
        fig, ax = pplt.subplots()
        ax = plot2d(
            _f, coords=[coords[axis_view[0]], coords[axis_view[1]]], ax=ax, **plot_kws
        )
        ax.format(xlabel=dims_units[axis_view[0]], ylabel=dims_units[axis_view[1]])
        plt.show()

    # Pass key word arguments for `update`.
    kws = dict()
    if cmap is not None:
        kws["cmap"] = cmap
    if profiles_checkbox:
        kws["profiles_checkbox"] = _profiles_checkbox
    kws["log"] = log
    kws["dim1"] = dim1
    kws["dim2"] = dim2
    if thresh_slider:
        kws["thresh_slider"] = thresh_slider
    for i, check in enumerate(checks, start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(sliders, start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)


def proj1d_interactive_slice(
    f,
    coords=None,
    default_ind=0,
    slice_type="int",
    dims=None,
    units=None,
    fig_kws=None,
    **plot_kws,
):
    """1D partial projection with interactive slicing.

    Parameters
    ----------
    f : ndarray
        A d-dimensional image.
    coords : list[ndarray]
        Grid coordinates for each dimension.
    default_ind : int
        Default index to plot.
    slice_type : {'int', 'range'}
        Whether to slice one index along the axis or a range of indices.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    kind : {'bar', 'line'}
        The kind of plot to draw.
    fig_kws : dict
        Key word arguments passed to `proplot.subplots`.
    **plot_kws
        Key word arguments passed to 1D plotting function.

    Returns
    -------
    gui : ipywidgets.widgets.interaction.interactive
        This widget can be displayed by calling `IPython.display.display(gui)`.
    """
    if coords is None:
        coords = [np.arange(f.shape[k]) for k in range(f.ndim)]
    if dims is None:
        dims = [f"x{i + 1}" for i in range(f.ndim)]
    if units is None:
        units = f.ndim * [""]
    dims_units = []
    for dim, unit in zip(dims, units):
        dims_units.append(f"{dim}" + f" [{unit}]" if unit != "" else dim)
    if fig_kws is None:
        fig_kws = dict()
    fig_kws.setdefault("figsize", (4.5, 1.5))
    plot_kws.setdefault("color", "black")
    plot_kws.setdefault("kind", "stepfilled")

    # Widgets
    dim1 = widgets.Dropdown(options=dims, index=default_ind, description="dim")

    # Sliders
    sliders, checks = [], []
    for k in range(f.ndim):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=f.shape[k],
                value=f.shape[k] // 2,
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                value=(0, f.shape[k]),
                min=0,
                max=f.shape[k],
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("Invalid `slice_type`.")
        slider.layout.display = "none"
        sliders.append(slider)
        checks.append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide/show sliders based on checkboxes."""
        for k in range(f.ndim):
            # Hide elements for dimensions being plotted.
            valid = dims[k] != dim1.value
            disp = None if valid else "none"
            for element in [sliders[k], checks[k]]:
                element.layout.display = disp
            # Uncheck boxes for dimensions being plotted.
            if not valid and checks[k].value:
                checks[k].value = False
            # Make sliders respond to check boxes.
            if not checks[k].value:
                sliders[k].layout.display = "none"

    # Update the slider list automatically.
    for element in (dim1, *checks):
        element.observe(hide, names="value")
    # Initial hide
    for k in range(f.ndim):
        if k == default_ind:
            checks[k].layout.display = "none"
            sliders[k].layout.display = "none"

    def update(**kws):
        """Update the figure."""
        dim1 = kws["dim1"]
        ind, checks = [], []
        for i in range(100):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                ind.append(kws[f"slider{i}"])
                
        # Return nothing if input does not make sense.
        for dim, check in zip(dims, checks):
            if check and dim == dim1:
                return
                
        # Slice, then project onto the specified axis.
        axis_view = dims.index(dim1)
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        for k in range(f.ndim):
            if type(ind[k]) is int:
                ind[k] = (ind[k], ind[k] + 1)
        ind = [ind[k] for k in axis_slice]
        idx = psdist.image.slice_idx(f.ndim, axis_slice, ind)
        profile = psdist.image.project(f[idx], axis_view)
        if np.max(profile) > 0:
            profile = profile / np.sum(profile)
            
        # Plot the projection.
        fig, ax = pplt.subplots(**fig_kws)
        ax.format(xlabel=dims_units[axis_view])
        vis.plot1d(coords[axis_view], profile, ax=ax, **plot_kws)
        plt.show()

    kws = {"dim1": dim1}
    for i, check in enumerate(checks, start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(sliders, start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)



"""Plotting routines for point data."""
from ipywidgets import interactive
from ipywidgets import widgets
from matplotlib import pyplot as plt
import numpy as np
import proplot as pplt

import psdist.cloud
import psdist.image
import psdist.visualization.image as vis_image
import psdist.visualization.visualization as vis


def auto_limits(X, sigma=None, pad=0.0, zero_center=False, share_xy=False):
    """Determine axis limits from coordinate array.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinate array for n points in d-dimensional space.
    sigma : float
        If a number is provided, it is used to set the limits relative to
        the standard deviation of the distribution.
    pad : float
        Fractional padding to apply to the limits.
    zero_center : bool
        Whether to center the limits on zero.
    share_xy : bool
        Whether to share x/y limits (and x'/y').

    Returns
    -------
    mins, maxs : list
        The new limits.
    """
    if sigma is None:
        mins = np.min(X, axis=0)
        maxs = np.max(X, axis=0)
    else:
        means = np.mean(X, axis=0)
        stds = np.std(X, axis=0)
        widths = 2.0 * sigma * stds
        mins = means - 0.5 * widths
        maxs = means + 0.5 * widths
    deltas = 0.5 * np.abs(maxs - mins)
    padding = deltas * pad
    mins = mins - padding
    maxs = maxs + padding
    if zero_center:
        maxs = np.max([np.abs(mins), np.abs(maxs)], axis=0)
        mins = -maxs
    if share_xy:
        widths = np.abs(mins - maxs)
        for (i, j) in [[0, 2], [1, 3]]:
            delta = 0.5 * (widths[i] - widths[j])
            if delta < 0.0:
                mins[i] -= abs(delta)
                maxs[i] += abs(delta)
            elif delta > 0.0:
                mins[j] -= abs(delta)
                maxs[j] += abs(delta)
    return [(_min, _max) for _min, _max in zip(mins, maxs)]


def plot_rms_ellipse(X, ax=None, level=1.0, center_at_mean=True, **ellipse_kws):
    """Compute and plot RMS ellipse from bunch coordinates.

    Parameters
    ----------
    X : ndarray, shape (n, 2)
        Coordinate array for n points in 2-dimensional space.
    ax : Axes
        The axis on which to plot.
    level : number of list of numbers
        If a number, plot the rms ellipse inflated by the number. If a list
        of numbers, repeat for each number.
    center_at_mean : bool
        Whether to center the ellipse at the image centroid.
    """
    center = np.mean(X, axis=0)
    if center_at_mean:
        center = (0.0, 0.0)
    Sigma = np.cov(X.T)
    return vis.rms_ellipse(Sigma, center, level=level, ax=ax, **ellipse_kws)


def scatter(X, ax=None, samples=None, **kws):
    """Convenience function for 2D scatter plot.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinate array for n points in d-dimensional space.
    ax : Axes
        The axis on which to plot.
    samples : int
        Plot this many random samples.
    **kws
        Key word arguments passed to `ax.scatter`.
    """
    if "color" in kws:
        kws["c"] = kws.pop("color")
    for kw in ["size", "ms"]:
        if kw in kws:
            kws["s"] = kws.pop(kw)
    kws.setdefault("c", "black")
    kws.setdefault("ec", "None")
    kws.setdefault("s", 2.0)
    _X = X
    if samples:
        _X = psdist.cloud.downsample(X, samples)
    return ax.scatter(_X[:, 0], _X[:, 1], **kws)


def hist(X, ax=None, bins="auto", limits=None, **kws):
    """Convenience function for 2D histogram with auto-binning.

    Parameters
    ----------
    X : ndarray, shape (n, 2)
        Coordinate array for n points in 2-dimensional space.
    ax : Axes
        The axis on which to plot.
    limits, bins : see `psdist.bunch.histogram`.
    **kws
        Key word arguments passed to `plotting.image`.
    """
    f, coords = psdist.cloud.histogram(X, bins=bins, limits=limits, centers=True)
    return vis_image.plot2d(f, coords=coords, ax=ax, **kws)


def kde(X, ax=None, coords=None, res=100, kde_kws=None, **kws):
    """Plot kernel density estimation (KDE).

    Parameters
    ----------
    X : ndarray, shape (n, 2)
        Coordinate array for n points in 2-dimensional space.
    ax : Axes
        The axis on which to plot.
    coords : [xcoords, ycoords]
        Coordinates along each axis of a two-dimensional regular grid on which to
        evaluate the density.
    res : int
        If coords is not provided, determines the evaluation grid resolution.
    kde_kws : dict
        Key word arguments passed to `psdist.bunch.kde`.
    **kws
        Key word arguments passed to `psdist.visualization.image.plot2`.
    """
    if kde_kws is None:
        kde_kws = dict()
    if coords is None:
        lb = np.min(X, axis=0)
        ub = np.max(X, axis=0)
        coords = [np.linspace(l, u, res) for l, u in zip(lb, ub)]
    estimator = psdist.cloud.gaussian_kde(X, **kde_kws)
    density = estimator.evaluate(psdist.image.get_grid_coords(*coords).T)
    density = np.reshape(density, [len(c) for c in coords])
    return vis_image.plot2d(density, coords=coords, ax=ax, **kws)


def plot2d(X, kind="hist", rms_ellipse=False, rms_ellipse_kws=None, ax=None, **kws):
    """Plot

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinate array for n points in d-dimensional space.
    kind : {'hist', 'contour', 'contourf', 'scatter', 'kde'}
        The kind of plot.
    rms_ellipse : bool
        Whether to plot the RMS ellipse.
    rms_ellipse_kws : dict
        Key word arguments passed to `visualization.cloud.plot_rms_ellipse`.
    ax : Axes
        The axis on which to plot.
    **kws
        Key word arguments passed to `visualization.cloud.plot2d`.
    """
    if kind == "hist":
        kws.setdefault("mask", True)
    func = None
    if kind in ["hist", "contour", "contourf"]:
        func = hist
        if kind in ["contour", "contourf"]:
            kws["kind"] = kind
    elif kind == "scatter":
        func = scatter
    elif kind == "kde":
        func = kde
    else:
        raise ValueError("Invalid plot kind.")
    _out = func(X, ax=ax, **kws)
    if rms_ellipse:
        if rms_ellipse_kws is None:
            rms_ellipse_kws = dict()
        plot_rms_ellipse(X, ax=ax, **rms_ellipse_kws)
    return _out


def joint(X, grid_kws=None, marg_hist_kws=None, marg_kws=None, **kws):
    """Joint plot.

    This is a convenience function; see `psdist.visualization.grid.JointGrid`.

    Parameters
    ----------
    X : ndarray, shape (n, 2)
        Coordinates of n points in 2-dimensional space.
    grid_kws : dict
        Key word arguments passed to `JointGrid`.
    marg_hist_kws : dict
        Key word arguments passed to `np.histogram` for 1D histograms.
    marg_kws : dict
        Key word arguments passed to `visualization.plot1d`.
    **kws
        Key word arguments passed to `visualization.image.plot2d.`

    Returns
    -------
    psdist.visualization.grid.JointGrid
    """
    from psdist.visualization.grid import JointGrid

    if grid_kws is None:
        grid_kws = dict()
    grid = JointGrid(**grid_kws)
    grid.plot_cloud(X, marg_hist_kws=marg_hist_kws, marg_kws=marg_kws, **kws)
    return grid


def corner(
    X,
    grid_kws=None,
    limits=None,
    labels=None,
    autolim_kws=None,
    prof_edge_only=False,
    update_limits=True,
    diag_kws=None,
    **kws,
):
    """Corner plot (scatter plot matrix).

    This is a convenience function; see `psdist.visualization.grid.CornerGrid`.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    limits : list[tuple], length n
        The (min, max) plot limits for each axis.
    labels : list[str], length n
        The axis labels.
    prof_edge_only : bool
        If plotting profiles on top of images (on off-diagonal subplots), whether
        to plot x profiles only in bottom row and y profiles only in left column.
    update_limits : bool
        Whether to extend the existing plot limits.
    diag_kws : dict
        Key word argument passed to `visualization.plot1d`.
    **kws
        Key word arguments pass to `visualization.cloud.plot2d`

    Returns
    -------
    psdist.visualization.grid.CornerGrid
        The `CornerGrid` on which the plot was drawn.
    """
    from psdist.visualization.grid import CornerGrid

    if grid_kws is None:
        grid_kws = dict()
    grid = CornerGrid(d=X.shape[1], **grid_kws)
    if labels is not None:
        grid.set_labels(labels)
    grid.plot_cloud(
        X,
        limits=limits,
        autolim_kws=autolim_kws,
        prof_edge_only=prof_edge_only,
        update_limits=update_limits,
        diag_kws=diag_kws,
        **kws,
    )
    return grid


def proj2d_interactive_slice(
    data=None,
    limits=None,
    share_limits=2,
    default_ind=(0, 1),
    slice_type="int",
    plot_res=64,
    slice_res=16,
    dims=None,
    units=None,
    options=None,
    autolim_kws=None,
    fig_kws=None,
    **plot_kws,
):
    """2D partial projection of one or more clouds (or series of clouds) with interactive slicing.

    Parameters
    ----------
    data : ndarray, shape (n, d) or list[ndarray] or list[list[ndarray]]
        - Coordinates of n points in d-dimensional space.
        - List of L clouds: generates widget to select the frame to plot.
        - K lists of L clouds: generates K-column figure with widget to select one
          of the L frames.
          Example: Compare the evolution of K=3 bunches at L=6 frames.
    limits : list[(min, max)]
        Limits along each axis.
    share_limits : 0, 1, 2
        Whether to share axis limits across frames. If 0, don't share. If 1, share
        between each subplot for each figure (frame). If 2, share for all 
        figures/subplots.
    default_ind : (int, int)
        Default view axis.
    slice_type : {"int", "range"}
        Whether to slice one index along the axis or a range of indices.
    plot_res, slice_res : int
        Default grid resolution for plotting/slicing. These can be updated using
        the interactive widgets.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    options : dict
        Determines the widgets to be displayed. Options are:
        - "auto_plot_res": automatically select plot resolution.
        - "discrete": discrete colormap norm. (Default: False).
        - "ellipse": plot rms ellipse. (Default: False)
        - "log": logarithmic colormap scaling. (Default: True)
        - "normalize": normalize x-px, y-py, z-pz to unit covariance matrix. (Default: False)
        - "profiles": plot profiles (line-outs) on bottom and left spines.
    autolim_kws : dict
        Key word arguments passed to `auto_limits`.
    fig_kws : dict
        Key word arguments passed to `proplot.subplots`.
    **plot_kws
        Key word arguments passed to `plot2d`.
    """
    # TODO: 
    # - clean up
    
    if type(data) is not list:
        data = [data]
    if type(data[0]) is not list:
        data = [data]

    n_rows = len(data)
    n_cols = len(data[0])
    n_dims = data[0][0].shape[1]

    for i in range(n_rows):
        if len(data[i]) != n_cols:
            raise ValueError("lists must have same length")

    for i in range(n_rows):
        for j in range(n_cols):
            if data[i][j].shape[1] != n_dims:
                raise ValueError("data must have the same number of dimensions.")

    if fig_kws is None:
        fig_kws = dict()
    plot_kws.setdefault("kind", "hist")
    plot_kws.setdefault(
        "rms_ellipse_kws",
        {
            "level": [1.0, 2.0, 3.0, 4.0, 5.0],
            "color": "white",
            "alpha": 0.2,
            "lw": 0.4,
        },
    )        
        
    # Compute limits [(xmin, xmax), ...] for each data[i][j].
    if autolim_kws is None:
        autolim_kws = dict()
    if limits is None:
        n_rows = len(data)
        n_cols = len(data[0])
        limits_list = []
        for i in range(n_rows):
            limits_list.append([])
            for j in range(n_cols):
                limits = auto_limits(data[i][j], **autolim_kws)
                limits_list[i].append(limits)
        limits_list = np.array(limits_list)
        if share_limits == 1:
            for j in range(n_cols):
                mins = np.min(limits_list[:, j, :, 0], axis=0)
                maxs = np.max(limits_list[:, j, :, 1], axis=0)
                limits = [(mins[k], maxs[k]) for k in range(n_dims)]
                for i in range(n_rows):
                    limits_list[i, j] = limits
        elif share_limits == 2:
            mins = np.min(limits_list[:, :, :, 0], axis=(0, 1))
            maxs = np.max(limits_list[:, :, :, 1], axis=(0, 1))
            limits = [(mins[k], maxs[k]) for k in range(n_dims)]
            for i in range(n_rows):
                for j in range(n_cols):
                    limits_list[i, j] = limits
    else:
        limits_list = [[limits for _ in range(n_cols)] for _ in range(n_rows)]
                
    # Set axis labels.
    if dims is None:
        dims = [f"x{i + 1}" for i in range(n_dims)]
    if units is None:
        units = n_dims * [""]
    dims_units = []
    for dim, unit in zip(dims, units):
        dims_units.append(f"{dim}" + f" [{unit}]" if unit != "" else dim)

    # Widgets
    _widgets = dict()
    _widgets["dim1"] = widgets.Dropdown(options=dims, index=default_ind[0], description="dim 1")
    _widgets["dim2"] = widgets.Dropdown(options=dims, index=default_ind[1], description="dim 2")
    _widgets["frame"] = widgets.BoundedIntText(min=0, max=(n_cols - 1), description="frame")
    _widgets["slice_res"] = widgets.BoundedIntText(
        value=slice_res,
        min=2,
        max=200,
        step=1,
        description="slice_res",
    )
    _widgets["plot_res"] = widgets.BoundedIntText(
        value=plot_res,
        min=2,
        max=350,
        step=1,
        description="plot_res",
    )
    _widgets["auto_plot_res"] = widgets.Checkbox(
        description="auto_plot_res", value=False
    )
    _widgets["log"] = widgets.Checkbox(description="log", value=False)
    _widgets["normalize"] = widgets.Checkbox(description="normalize", value=False)
    _widgets["discrete"] = widgets.Checkbox(description="discrete", value=False)
    _widgets["ellipse"] = widgets.Checkbox(description="ellipse", value=False)
    _widgets["profiles"] = widgets.Checkbox(description="profiles", value=False)

    # Sliders and checkboxes for slicing:
    _widgets["sliders"] = []
    _widgets["checks"] = []
    for k in range(n_dims):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=int(_widgets["slice_res"].value / 2),
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=(0, _widgets["slice_res"].value - 1),
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("Invalid `slice_type`.")
        slider.layout.display = "none"
        _widgets["sliders"].append(slider)
        _widgets["checks"].append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide inactive sliders."""
        for k in range(n_dims):
            # Hide elements for dimensions being plotted.
            valid = dims[k] not in [_widgets["dim1"].value, _widgets["dim2"].value]
            disp = None if valid else "none"
            for element in [_widgets["sliders"][k], _widgets["checks"][k]]:
                element.layout.display = disp
            # Uncheck boxes for dimensions being plotted.
            if not valid and _widgets["checks"][k].value:
                _widgets["checks"][k].value = False
            # Make sliders respond to check boxes.
            if not _widgets["checks"][k].value:
                _widgets["sliders"][k].layout.display = "none"
            _widgets["plot_res"].layout.display = (
                "none" if _widgets["auto_plot_res"].value else None
            )

    # Make slider visiblity depend on checkmarks.
    for element in (
        _widgets["dim1"],
        _widgets["dim2"],
        *_widgets["checks"],
        _widgets["auto_plot_res"],
    ):
        element.observe(hide, names="value")

    # Initial hide
    for k in range(n_dims):
        if k in default_ind:
            _widgets["checks"][k].layout.display = "none"
            _widgets["sliders"][k].layout.display = "none"
    if n_cols == 1:
        _widgets["frame"].layout.display = "none"

    # Set default options.
    if options is None:
        options = dict()
    options.setdefault("auto_plot_res", False)
    options.setdefault("discrete", False)
    options.setdefault("ellipse", False)
    options.setdefault("log", True)
    options.setdefault("normalize", False)
    options.setdefault("profiles", False)

    # Show/hide widgets based on `options`.
    for name, setting in options.items():
        _widgets[name].layout.display = None if setting else "none"

    def update(**kws):
        # Collect key word arguments.
        frame = kws["frame"]
        dim1 = kws["dim1"]
        dim2 = kws["dim2"]
        slice_res = kws["slice_res"]
        plot_res = kws["plot_res"]
        auto_plot_res = kws["auto_plot_res"]

        # Update the slider ranges/values based on slice_res.
        for slider in _widgets["sliders"]:
            slider.max = slice_res - 1

        # Collect slice indices.
        ind, checks = [], []
        for i in range(1, n_dims + 1):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                _ind = kws[f"slider{i}"]
                if type(_ind) is int:
                    _ind = (_ind, _ind + 1)
                ind.append(_ind)

        # Exit if input is invalid.
        for dim, check in zip(dims, checks):
            if check and dim in (dim1, dim2):
                return
        if dim1 == dim2:
            return
        
        # Collect data.        
        _data = [data[index][frame] for index in range(n_rows)]
        _limits_list = limits_list[:, frame, :, :]
        
        # Normalize coordinates.
        if kws["normalize"]:
            _data = [psdist.cloud.norm_xxp_yyp_zzp(_X, scale_emittance=True) for _X in _data]
            _limits_list = [auto_limits(_X, **autolim_kws) for _X in _data]
            if share_limits > 0:
                _limits_list = [vis.stack_limits(_limits_list) for _ in range(len(_data))]

        # Slice.
        axis_view = [dims.index(dim) for dim in (dim1, dim2)]
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        if axis_slice:
            for index in range(n_rows):
                slice_limits = []
                for k in axis_slice:
                    (imin, imax) = ind[k]
                    (xmin, xmax) = _limits_list[index][k]
                    edges = np.linspace(xmin, xmax, slice_res + 1)
                    if imax > len(edges) - 1:
                        print(f"{dims[k]} out of range.")
                        return
                    slice_limits.append((edges[imin], edges[imax]))
                _data[index] = psdist.cloud.slice_planar(_data[index], axis=axis_slice, limits=slice_limits)
                
        # Handle empty slice (do nothing).
        for _X in _data:
            if _X.shape[0] == 0:
                return
                                                                                
        # Update plotting key word arguments.
        if plot_kws["kind"] != "scatter":
            plot_kws["bins"] = "auto" if auto_plot_res else plot_res
            plot_kws["norm"] = "log" if kws["log"] else None
            plot_kws["discrete"] = kws["discrete"]
            plot_kws["rms_ellipse"] = kws["ellipse"]
            plot_kws["profx"] = kws["profiles"]
            plot_kws["profy"] = kws["profiles"]

            # Temporary bug fix: If we check and then uncheck "log", and
            # the colorbar has minor ticks, the tick label formatter will
            # remain in "log" mode forever after.
            if "colorbar_kw" in plot_kws:
                if "tickminor" in plot_kws["colorbar_kw"] and not kws["log"]:
                    plot_kws["colorbar_kw"]["formatter"] = None
                            
        # Plot the selected points.
        sharex = (share_limits and n_rows)
        fig, axs = pplt.subplots(ncols=n_rows, sharex=sharex, sharey=sharex, **fig_kws)
        for index, ax in enumerate(axs):
            limits = [_limits_list[index][k] for k in axis_view]
            if plot_kws["kind"] != "scatter":
                plot_kws["limits"] = limits
            plot2d(_data[index][:, axis_view], ax=ax, **plot_kws)
            ax.format(xlim=limits[0], ylim=limits[1])
        axs.format(
            xlabel=dims_units[axis_view[0]], 
            ylabel=dims_units[axis_view[1]],
        )
        plt.show()

    # Pass key word arguments to `ipywidgets.interactive`.
    kws = dict()
    kws["frame"] = _widgets["frame"]
    kws["dim1"] = _widgets["dim1"]
    kws["dim2"] = _widgets["dim2"]
    kws["slice_res"] = _widgets["slice_res"]
    kws["plot_res"] = _widgets["plot_res"]
    kws["auto_plot_res"] = _widgets["auto_plot_res"]
    kws["log"] = _widgets["log"]
    kws["discrete"] = _widgets["discrete"]
    kws["ellipse"] = _widgets["ellipse"]
    kws["normalize"] = _widgets["normalize"]
    kws["profiles"] = _widgets["profiles"]
    for i, check in enumerate(_widgets["checks"], start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(_widgets["sliders"], start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)



"""Functions for point clouds."""
import collections

import numpy as np
import scipy.interpolate
import scipy.special
import scipy.stats

from psdist import ap
import psdist.utils as utils
from psdist.utils import array_like
from psdist.utils import centers_from_edges
from psdist.utils import cov2corr
from psdist.utils import random_selection


# Analysis
# ------------------------------------------------------------------------------
def mean(X):
    """Compute mean (centroid).

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.

    Returns
    -------
    ndarray, shape (d,)
        The centroid coordinates.
    """
    return np.mean(X, axis=0)


def cov(X):
    """Compute covariance matrix.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.

    Returns
    -------
    ndarray, shape (d, d)
        The covariance matrix of second-order moments.
    """
    return np.cov(X.T)


def corr(X):
    """Compute correlation matrix.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.

    Returns
    -------
    ndarray, shape (d, d)
        The correlation matrix.
    """
    return cov2corr(np.cov(X.T))


def get_radii(X):
    return np.linalg.norm(X, axis=1)


def get_ellipsoid_radii(X):
    Sigma_inv = np.linalg.inv(np.cov(X.T))
    func = lambda point: np.sqrt(np.linalg.multi_dot([point.T, Sigma_inv, point]))
    return transform(X, func)


def enclosing_sphere(X, axis=None, fraction=1.0):
    """Scales sphere until it contains some fraction of points.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        The distribution is projected onto this axis before proceeding. The
        ellipsoid is defined in this subspace.
    fraction : float
        Fraction of points in sphere.

    Returns
    -------
    radius : float
        The sphere radius.
    """
    radii = np.sort(get_radii(project(X, axis)))
    index = int(np.round(X.shape[0] * fraction)) - 1
    return radii[index]


def enclosing_ellipsoid(X, axis=None, fraction=1.0):
    """Scale the rms ellipsoid until it contains some fraction of points.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        The distribution is projected onto this axis before proceeding. The
        ellipsoid is defined in this subspace.
    fraction : float
        Fraction of points enclosed.

    Returns
    -------
    float
        The ellipsoid "radius" (x^T Sigma^-1 x) relative to the rms ellipsoid.
    """
    radii = np.sort(get_ellipsoid_radii(project(X, axis)))
    index = int(np.round(X.shape[0] * fraction)) - 1
    return radii[index]


## Statistical distance measures (https://journals.aps.org/pre/abstract/10.1103/PhysRevE.106.065302)
## [...]


# Transformation
# ------------------------------------------------------------------------------


def project(X, axis=None):
    """Axis-aligned projection. (Just calls `X[:, axis]`.)

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple[int], length l
        The axis on which to project the points.

    Returns
    -------
    ndarray, shape (n, l)
        The points projected onto the specified axis.
    """
    if axis is None:
        axis = tuple(np.arange(X.shape[1]))
    if array_like(axis) and len(axis) > X.shape[1]:
        raise ValueError("Invalid projection axis.")
    return X[:, axis]


def transform(X, func=None, **kws):
    """Apply a nonlinear transformation.

    This function just calls `np.apply_along_axis`.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    function : callable
        Function applied to each point in X. Call signature is
        `function(point, **kws)` where `point` is an n-dimensional
        point given by one row of `X`.
    **kws
        Key word arguments for

    Returns
    -------
    ndarray, shape (n, d)
        The transformed distribution.
    """
    return np.apply_along_axis(lambda point: func(point, **kws), 1, X)


def transform_linear(X, M):
    """Apply a linear transformation.

    This function just calls `np.apply_along_axis`.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    M : ndarray, shape (d, d)
        A linear transfer matrix.

    Returns
    -------
    ndarray, shape (n, d)
        The transformed distribution.
    """
    func = lambda point: np.matmul(M, point)
    return transform(X, lambda point: np.matmul(M, point))


def shift(X, delta=0.0):
    return X + delta


def scale(X, factor=1.0):
    return X * factor


def slice_planar(X, axis=None, center=None, width=None, limits=None):
    """Return points within a planar slice.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        Slice axes. For example, (0, 1) will slice along the first and
        second axes of the array.
    center : ndarray, shape (n,)
        The center of the box.
    width : ndarray, shape (d,)
        The width of the box along each axis.
    limits : ndarray, shape (d, 2)
        The (min, max) along each axis. Overrides `center` and `edges` if provided.

    Returns
    -------
    ndarray, shape (?, n)
        The points within the box.
    """
    n, d = X.shape
    if not array_like(axis):
        axis = (axis,)
    if limits is None:
        if not array_like(center):
            center = np.full(d, center)
        if not array_like(width):
            width = np.full(d, width)
        center = np.array(center)
        width = np.array(width)
        limits = list(zip(center - 0.5 * width, center + 0.5 * width))  
    limits = np.array(limits)
    if limits.ndim == 1:
        limits = limits[None, :]
    conditions = []
    for j, (umin, umax) in zip(axis, limits):
        conditions.append(X[:, j] > umin)
        conditions.append(X[:, j] < umax)
    idx = np.logical_and.reduce(conditions)
    return X[idx, :]


def slice_sphere(X, axis=None, rmin=0.0, rmax=None):
    """Return points within a spherical shell slice.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        The subspace in which to define the sphere.
    rmin, rmax : float
        Inner/outer radius of spherical shell.

    Returns
    -------
    ndarray, shape (?, d)
        The points within the sphere.
    """
    if rmax is None:
        rmax = np.inf
    radii = get_radii(project(X, axis))
    idx = np.logical_and(radii > rmin, radii < rmax)
    return X[idx, :]


def slice_ellipsoid(X, axis=None, rmin=0.0, rmax=None):
    """Return points within an ellipsoidal shell slice.

    The ellipsoid is defined by the covariance matrix of the
    distribution.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        The subspace in which to define the ellipsoid.
    rmin, rmax : list[float]
        Min/max "radius" (x^T Sigma^-1 x). relative to covariance matrix.

    Returns
    -------
    ndarray, shape (?, d)
        Points within the shell.
    """
    if rmax is None:
        rmax = np.inf
    radii = get_ellipsoid_radii(project(X, axis))
    idx = np.logical_and(rmin < radii, radii < rmax)
    return X[idx, :]


def slice_contour(X, axis=None, lmin=0.0, lmax=1.0, interp=True, **hist_kws):
    """Return points within a contour shell slice.

    The slice is defined by the density contours in the subspace defined by
    `axis`.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    axis : tuple
        The subspace in which to define the density contours.
    lmin, lmax : list[float]
        If `f` is the density in the subspace defined by `axis`, then we select
        points where lmin <= f / max(f) <= lmax.
    interp : bool
        If True, compute the histogram, then interpolate and evaluate the
        resulting function at each point in `X`. Otherwise we keep track
        of the indices in which each point lands when it is binned,
        and accept the point if it's bin has a value within fmin and fmax.
        The latter is a lot slower.

    Returns
    -------
    ndarray, shape (?, d)
        Points within the shell.
    """
    _X = project(X, axis)
    hist, edges = histogram(_X, **hist_kws)
    hist = hist / np.max(hist)
    centers = [0.5 * (e[:-1] + e[1:]) for e in edges]
    if interp:
        fint = scipy.interpolate.RegularGridInterpolator(
            centers,
            hist,
            method="linear",
            bounds_error=False,
            fill_value=0.0,
        )
        values = fint(_X)
        idx = np.logical_and(lmin <= values, values <= lmax)
    else:
        valid_indices = np.vstack(
            np.where(np.logical_and(lmin <= hist, hist <= lmax))
        ).T
        indices = np.vstack(
            [np.digitize(_X[:, k], edges[k]) for k in range(_X.shape[1])]
        ).T
        idx = []
        for i in range(len(indices)):
            if indices[i].tolist() in valid_indices.tolist():
                idx.append(i)
    return X[idx, :]


def norm_xxp_yyp_zzp(X, scale_emittance=False):
    """Normalize x-px, y-py, z-pz, ...

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional phase space (d is even).
    scale_emittance : bool
        Whether to divide the coordinates by the square root of the rms emittance.

    Returns
    -------
    Xn : ndarray, shape (n, d)
        Normalized phase space coordinate array.
    """
    if X.shape[1] % 2 != 0:
        raise ValueError("X must have an even number of columns.")
    Sigma = np.cov(X.T)
    Xn = np.zeros(X.shape)
    for i in range(0, X.shape[1], 2):
        sigma = Sigma[i : i + 2, i : i + 2]
        alpha, beta = ap.twiss(sigma)
        Xn[:, i] = X[:, i] / np.sqrt(beta)
        Xn[:, i + 1] = (np.sqrt(beta) * X[:, i + 1]) + (alpha * X[:, i] / np.sqrt(beta))
        if scale_emittance:
            eps = ap.apparent_emittance(sigma)
            Xn[:, i : i + 2] = Xn[:, i : i + 2] / np.sqrt(eps)
    return Xn


def decorrelate(X):
    """Remove cross-plane correlations by permuting (x, x'), (y, y'), (z, z') pairs.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of k points in d-dimensional space (d is even).

    Returns
    -------
    ndarray, shape (n, d)
        The decorrelated coordinates.
    """
    if X.shape[1] % 2 != 0:
        raise ValueError("X must have even number of columns.")
    for i in range(0, X.shape[1], 2):
        idx = np.random.permutation(np.arange(X.shape[0]))
        X[:, i : i + 2] = X[idx, i : i + 2]
    return X


def downsample(X, samples):
    """Select a random subset of points.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    samples : int or float
        The number of samples to keep If less than 1, specifies
        the fraction of points.

    Returns
    -------
    ndarray, shape (<= n, d)
        The downsampled coordinate array.
    """
    samples = min(samples, X.shape[0])
    idx = random_selection(np.arange(X.shape[0]), samples)
    return X[idx, :]


# Density estimation
# ------------------------------------------------------------------------------


def histogram_bin_edges(X, bins=10, limits=None):
    """Multi-dimensional histogram bin edges.

    This function calls `np.histogram_bin_edges` along each axis of X. 
    
    See https://numpy.org/doc/stable/reference/generated/numpy.histogram_bin_edges.html

    Parameters
    ----------
    bins : int or str    
        If `bins` is an int, it defines the number of equal-width bins in the 
        given range (10, by default). 

        If `bins` is a string, `histogram_bin_edges` will use the method chosen 
        to calculate the optimal number of bins.

        if `bins` is a sequence of floats, it defines the bin edges, including 
        the rightmost edge.

        A list of {str / int / float sequence} may be provided such that bins[i] 
        corresponds to axis i.
    limits : (float, float)
        The lower and upper range of the bins.  If not provided, the limits are 
        ``[(np.min(X[:, i]), np.max(X[:, i])) for i in range(X.shape[1])]``.

    Returns
    -------
    edges : list[ndarray]
        Bin edges along each axis.
    """
    if X.ndim == 1:
        return np.histogram_bin_edges(X, bins, limits)
    # `[2, 3, 4, 5]` could mean "2 bins along axis 0, 3 bins along axis 1, ..." 
    # or "bin edges [2.0, 3.0, 4.0, 5.0] along each axis". We assume the 
    # former if `bins` is a sequence of int and the latter if `bins` is a
    # sequence of float.
    if array_like(bins) and type(bins[0]) is float:
        bins = X.shape[1] * [bins]
    # If a single int/str is provided, apply to all axes.
    if not array_like(bins):
        bins = X.shape[1] * [bins]
    # Same for `limits`. If a (min, max) tuple (or None) is provided, apply 
    # to all axes.
    if limits is None or (limits[0] is not None and not array_like(limits[0])):
        limits = X.shape[1] * [limits]
    return [
        np.histogram_bin_edges(X[:, i], bins[i], limits[i]) 
        for i in range(X.shape[1])
    ]


def histogram(X, bins=10, limits=None, centers=False):
    """Multi-dimensional histogram.

    Parameters
    ----------
    See `histogram_bin_edges`.

    Returns
    -------
    See `np.histogramdd`.
    """
    if X.ndim == 1:
        bins = np.histogram_bin_edges(X, bins, limits)
        hist, _ = np.histogram(X, bins=bins)
        if centers:
            bins = utils.centers_from_edges(bins)
        return hist, bins

    bins = histogram_bin_edges(X, bins=bins, limits=limits)
    hist, _ = np.histogramdd(X, bins)   
    if centers:
        bins = [utils.centers_from_edges(b) for b in bins]
    return hist, bins


def sparse_histogram(X, bins=10, limits=None, centers=False, eps=1.0e-12):
    """Compute sparse multidimensional histogram.
    
    Parameters
    ----------
    Same as `histogram`.
    eps : float
        Small constant added to last bin edge to work around np.digitize. (See 
        comment in code.)

    Returns
    ------
    indices : ndarray, shape (k, d)
        Indices of nonzero bins in d-dimensional histogram.
    counts : ndarray, shape (k,)
        Counts of nonzero bins in d-dimensional histogram.
    bins : list(ndarray)
        List of bin edges or centers along each axis.
    """
    d = X.shape[1]
    bins = histogram_bin_edges(X, bins=bins, limits=limits)
    # Add a small number to the last bin edge. np.digitize includes computes index
    # i such that bins[i-1] <= x < bins[i], but we want bins[i-1] <= x <= bins[i].
    for axis in range(len(bins)):
        bins[axis][-1] = bins[axis][-1] + eps
    # Get multidimensional bin index of each point.
    indices = []
    for axis in range(d):
        idx = np.digitize(X[:, axis], bins[axis])
        # Drop points outside the bin range.
        idx = idx[np.logical_and(idx > 0, idx < len(bins[axis]))]
        # Subtract 1 so that 0 indexes the first bin.
        idx = idx - 1
        indices.append(idx)
    indices = np.vstack(indices).T
    # Convert to indices in flattened histogram. 
    shape = [len(b) - 1 for b in bins]
    indices = [np.ravel_multi_index(idx, shape) for idx in indices]
    # Count the indices/counts of each nonzero bin.
    counter = collections.Counter(indices)
    counts = np.array(list(counter.values()))
    indices = np.array(list(counter.keys()))
    # Convert from flat to multidimensional indices.
    indices = np.unravel_index(indices, shape)
    indices = np.vstack(indices).T
    # Return bin centers, rather than edges, if desired.
    if centers:
        bins = [centers_from_edges(b) for b in bins]
    return indices, counts, bins
    

def gaussian_kde(X, **kws):
    """Gaussian kernel density estimation (KDE).

    This function just calls `scipy.stats.gaussian_kde`.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    **kws
        Key word arguments

    Returns
    -------
    estimator : scipy.stats.gaussian_kde
        The density estimator.
    """
    return scipy.stats.gaussian_kde(X.T, **kws)


def radial_histogram(X, **kws):
    """Count number of points within spherical shells, with counts normalized by shell volume.

    Parameters
    ----------
    X : ndarray, shape (n, d)
        Coordinates of n points in d-dimensional space.
    **kws
        Key word arguments for `histogram`.
    """
    radii = get_radii(X)
    hist, bins = histogram(radii, **kws)
    if "centers" in kws and kws["centers"]:
        _edges = utils.edges_from_centers(bins)
    else:
        _edges = bins
    for i in range(len(_edges) - 1):
        rmin = _edges[i]
        rmax = _edges[i + 1]
        hist[i] = hist[i] / utils.volume_sphere_shell(rmin=rmin, rmax=rmax, n=X.shape[1])
    return hist, bins


import warnings

from ipywidgets import interact
from ipywidgets import interactive
from ipywidgets import widgets
from matplotlib import patches
from matplotlib import pyplot as plt
import numpy as np
import proplot as pplt
import scipy.optimize

import psdist.image
import psdist.cloud
import psdist.utils
import psdist.visualization.cloud as vis_cloud
import psdist.visualization.image as vis_image


def ellipse(c1=1.0, c2=1.0, angle=0.0, center=(0, 0), ax=None, **kws):
    """Plot ellipse with semi-axes `c1`,`c2` tilted `angle`radians below the x axis."""
    kws.setdefault("fill", False)
    kws.setdefault("color", "black")
    width = 2.0 * c1
    height = 2.0 * c2
    return ax.add_patch(
        patches.Ellipse(center, width, height, -np.degrees(angle), **kws)
    )


def circle(r=1.0, center=(0.0, 0.0), ax=None, **kws):
    """Plot a circle."""
    return ellipse(r, r, center=center, ax=ax, **kws)


def rms_ellipse_dims(Sigma, axis=(0, 1)):
    """Return dimensions of projected rms ellipse.

    Parameters
    ----------
    Sigma : ndarray, shape (2n, 2n)
        The phase space covariance matrix.
    axis : 2-tuple
        The axis on which to project the covariance ellipsoid. Example: if the
        axes are {x, xp, y, yp}, and axis=(0, 2), then the four-dimensional
        ellipsoid is projected onto the x-y plane.
    ax : plt.Axes
        The ax on which to plot.

    Returns
    -------
    c1, c2 : float
        The ellipse semi-axis widths.
    angle : float
        The tilt angle below the x axis [radians].
    """
    i, j = axis
    sii, sjj, sij = Sigma[i, i], Sigma[j, j], Sigma[i, j]
    angle = -0.5 * np.arctan2(2 * sij, sii - sjj)
    sin, cos = np.sin(angle), np.cos(angle)
    sin2, cos2 = sin**2, cos**2
    c1 = np.sqrt(abs(sii * cos2 + sjj * sin2 - 2 * sij * sin * cos))
    c2 = np.sqrt(abs(sii * sin2 + sjj * cos2 + 2 * sij * sin * cos))
    return c1, c2, angle


def rms_ellipse(Sigma=None, center=None, level=1.0, ax=None, **ellipse_kws):
    """Plot RMS ellipse from 2 x 2 covariance matrix."""
    if type(level) not in [list, tuple, np.ndarray]:
        level = [level]
    c1, c2, angle = rms_ellipse_dims(Sigma)
    for level in level:
        _c1 = c1 * level
        _c2 = c2 * level
        ellipse(_c1, _c2, angle=angle, center=center, ax=ax, **ellipse_kws)
    return ax


def fit_linear(x, y):
    """Return (yfit, slope, intercept) from linear fit."""

    def func(x, slope, intercept):
        return slope * x + intercept

    popt, pcov = scipy.optimize.curve_fit(func, x, y)
    slope, intercept = popt
    return func(x, slope, intercept), slope, intercept


def fit_normal(x, y):
    """Return (yfit, sigma, mu, amplitude, offset) from Gaussian fit."""

    def func(x, sigma, mu, amplitude, offset):
        amplitude = amplitude / (sigma * np.sqrt(2.0 * np.pi))
        return offset + amplitude * np.exp(-0.5 * ((x - mu) / sigma) ** 2)

    popt, pcov = scipy.optimize.curve_fit(func, x, y)
    sigma, mu, amplitude, offset = popt
    return func(x, sigma, mu, amplitude, offset), sigma, mu, amplitude, offset


def plot1d(x, y, ax=None, offset=0.0, flipxy=False, kind="line", **kws):
    """Convenience function for one-dimensional line/step/bar plots."""
    kws.setdefault("color", "black")

    func = ax.plot
    if kind in ["line", "step"]:
        if flipxy:
            func = ax.plotx
        else:
            func = ax.plot
        if kind == "step":
            kws.setdefault("drawstyle", "steps-mid")
    elif kind in ["linefilled", "stepfilled"]:
        if flipxy:
            func = ax.fill_betweenx
        else:
            func = ax.fill_between
        kws.setdefault("alpha", 1.0)
        if kind == "stepfilled":
            kws.setdefault("step", "mid")
    elif kind == "bar":
        if flipxy:
            func = ax.barh
        else:
            func = ax.bar

    # Handle offset
    if kind == "bar":
        kws["left" if flipxy else "bottom"] = offset * np.ones(len(x))
        return func(x, y, **kws)
    elif kind in ["linefilled", "stepfilled"]:
        return func(x, offset, y + offset, **kws)
    return func(x, y + offset, **kws)


def stack_limits(limits_list):
    limits_list = np.array(limits_list)
    mins = np.min(limits_list[:, :, 0], axis=0)
    maxs = np.max(limits_list[:, :, 1], axis=0)
    return [(mins[i], maxs[i]) for i in range(len(mins))]
